# sage_setup: distribution = sagemath-schemes
# distutils: sources = sage/modular/arithgroup/sl2z.cpp sage/modular/arithgroup/farey.cpp
# distutils: language = c++
# distutils: extra_compile_args = -std=c++17
# sage.doctest: needs sage.libs.pari
r"""
Farey symbol for arithmetic subgroups of `\PSL_2(\ZZ)`

AUTHORS:

- Hartmut Monien (08 - 2011)

based on the *KFarey* package by Chris Kurth. Implemented as C++ module
for speed.
"""

# ****************************************************************************
#       Copyright (C) 2011 Hartmut Monien <monien@th.physik.uni-bonn.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cpython.object cimport PyObject_RichCompare
from itertools import groupby
from cysignals.signals cimport sig_on, sig_off

from sage.libs.gmpxx cimport *

from sage.rings.real_mpfr import RR
from sage.rings.cc import CC
from sage.rings.integer cimport Integer
from sage.rings.infinity import infinity
from .congroup_gammaH import GammaH_class
from .congroup_gamma1 import Gamma1_class
from .congroup_gamma0 import Gamma0_class
from .congroup_gamma import Gamma_class
from .congroup_sl2z import SL2Z
from sage.modular.cusps import Cusp

from sage.misc.decorators import options
from sage.misc.cachefunc import cached_method
from sage.structure.richcmp cimport richcmp_not_equal


cdef extern from "sl2z.hpp":
    cppclass cpp_SL2Z "SL2Z":
        cpp_SL2Z(int, int, int, int)
        cpp_SL2Z(mpz_class, mpz_class, mpz_class, mpz_class)
        mpz_class a()
        mpz_class b()
        mpz_class c()
        mpz_class d()

cdef extern from "farey.hpp":
    cppclass is_element_Gamma0:
        is_element_Gamma0(int)
    cppclass is_element_Gamma1:
        is_element_Gamma1(int)
    cppclass is_element_Gamma:
        is_element_Gamma(int)
    cppclass is_element_GammaH:
        is_element_GammaH(int, object)
    cppclass cpp_farey "FareySymbol":
        cpp_farey()
        cpp_farey(object)
        cpp_farey(object, is_element_Gamma*)
        cpp_farey(object, is_element_Gamma0*)
        cpp_farey(object, is_element_Gamma1*)
        cpp_farey(object, is_element_GammaH*)
        size_t genus()
        size_t index()
        size_t level()
        size_t nu2()
        size_t nu3()
        object is_element(mpz_t, mpz_t, mpz_t, mpz_t)
        object word_problem(mpz_t, mpz_t, mpz_t, mpz_t, cpp_SL2Z *)
        size_t get_cusp_class(mpz_t, mpz_t)
        object get_cusps()
        object get_cusp_widths()
        object get_transformation_to_cusp(mpz_t, mpz_t)
        object get_fractions()
        object get_coset()
        object get_generators()
        object get_pairings()
        object get_paired_sides()
        object get_pairing_matrices()
        object dumps()


cdef class Farey:
    r"""
    A class for calculating Farey symbols of arithmetics subgroups of
    `\PSL_2(\ZZ)`.

    The arithmetic subgroup can be either any of
    the congruence subgroups implemented in Sage, i.e. Gamma, Gamma0,
    Gamma1 and GammaH or a subgroup of `\PSL_2(\ZZ)` which is
    given by a user written helper class defining membership in that
    group.

    REFERENCES:

    - Ravi S. Kulkarni, ''An arithmetic-geometric method in the study of the
      subgroups of the modular group'', `Amer. J. Math., 113(6):1053--1133,
      1991. <http://www.jstor.org/stable/2374900>`_

    INPUT:

    - ``G`` -- an arithmetic subgroup of `\PSL_2(\ZZ)`

    EXAMPLES:

    Create a Farey symbol for the group `\Gamma_0(11)`::

        sage: f = FareySymbol(Gamma0(11)); f
        FareySymbol(Congruence Subgroup Gamma0(11))

    Calculate the generators::

         sage: f.generators()
         [
         [1 1]  [ 7 -2]  [ 8 -3]  [-1  0]
         [0 1], [11 -3], [11 -4], [ 0 -1]
         ]

    Pickling the FareySymbol and recovering it::

         sage: f == loads(dumps(f))
         True

    Calculate the index of `\Gamma_H(33, [2, 5])` in
    `\PSL_2(\ZZ)` via FareySymbol::

         sage: FareySymbol(GammaH(33, [2, 5])).index()
         48

    Calculate the generators of `\Gamma_1(4)`::

         sage: FareySymbol(Gamma1(4)).generators()
         [
         [1 1]  [-3  1]
         [0 1], [-4  1]
         ]

    Calculate the generators of the :meth:`example
    <sage.modular.arithgroup.arithgroup_perm.HsuExample10>` of an
    index 10 arithmetic subgroup given by Tim Hsu::

         sage: from sage.modular.arithgroup.arithgroup_perm import HsuExample10         # needs sage.groups
         sage: FareySymbol(HsuExample10()).generators()                                 # needs sage.groups
         [
         [1 2]  [-2  1]  [ 4 -3]
         [0 1], [-7  3], [ 3 -2]
         ]

    Calculate the generators of the group `\Gamma' =
    \Gamma_0(8)\cap\Gamma_1(4)` using a helper class to define group membership::

         sage: class GPrime:
         ....:     def __contains__(self, M):
         ....:         return M in Gamma0(8) and M in Gamma1(4)

         sage: FareySymbol(GPrime()).generators()
         [
         [1 1]  [ 5 -1]  [ 5 -2]
         [0 1], [16 -3], [ 8 -3]
         ]

    Calculate cusps of arithmetic subgroup defined via permutation group::

        sage: # needs sage.groups
        sage: L = SymmetricGroup(4)('(1, 2, 3)')
        sage: R = SymmetricGroup(4)('(1, 2, 4)')
        sage: FareySymbol(ArithmeticSubgroup_Permutation(L, R)).cusps()
        [-1, Infinity]

    Calculate the left coset representation of `\Gamma_H(8, [3])`::

         sage: FareySymbol(GammaH(8, [3])).coset_reps()
         [
         [1 0]  [ 4 -1]  [ 3 -1]  [ 2 -1]  [ 1 -1]  [ 3 -1]  [ 2 -1]  [-1  0]
         [0 1], [ 1  0], [ 1  0], [ 1  0], [ 1  0], [ 4 -1], [ 3 -1], [ 3 -1],
         [ 1 -1]  [-1  0]  [ 0 -1]  [-1  0]
         [ 2 -1], [ 2 -1], [ 1 -1], [ 1 -1]
         ]
    """
    cdef cpp_farey *this_ptr
    cdef object group

    def __cinit__(self, group, data=None):
        r"""
        Initialize FareySymbol::

            sage: FareySymbol(Gamma0(23))
            FareySymbol(Congruence Subgroup Gamma0(23))
        """
        self.group = group
        # if data is present we want to restore
        if data is not None:
            sig_on()
            self.this_ptr = new cpp_farey(data)
            sig_off()
            return
        # to accelerate the calculation of the FareySymbol
        # we implement the tests for the standard congruence groups
        # in the c++ module. For a general group the test if an element
        # of SL2Z is in the group the python __contains__ attribute
        # of the group is called
        cdef int p
        if hasattr(group, "level"):
            p = group.level()
        if group == SL2Z:
            sig_on()
            self.this_ptr = new cpp_farey()
            sig_off()
        elif isinstance(group, Gamma0_class):
            sig_on()
            self.this_ptr = new cpp_farey(group, new is_element_Gamma0(p))
            sig_off()
        elif isinstance(group, Gamma1_class):
            sig_on()
            self.this_ptr = new cpp_farey(group, new is_element_Gamma1(p))
            sig_off()
        elif isinstance(group, Gamma_class):
            sig_on()
            self.this_ptr = new cpp_farey(group, new is_element_Gamma(p))
            sig_off()
        elif isinstance(group, GammaH_class):
            sig_on()
            l = group._GammaH_class__H
            self.this_ptr = new cpp_farey(group, new is_element_GammaH(p, l))
            sig_off()
        else:
            sig_on()
            self.this_ptr = new cpp_farey(group)
            sig_off()

    def __dealloc__(self):
        r"""
        Remove reference to FareySymbol.

        TESTS::

            sage: F = FareySymbol(Gamma0(23))
            sage: del F
        """
        del self.this_ptr

    @cached_method
    def pairing_matrices_to_tietze_index(self):
        r"""
        Obtain the translation table from pairing matrices
        to generators.

        The result is cached.

        OUTPUT:

        a list where the `i`-th entry is a nonzero integer `k`,
        such that if `k > 0` then the `i`-th pairing matrix is (up to sign)
        the `(k-1)`-th generator and, if `k < 0`, then the `i`-th pairing
        matrix is (up to sign) the inverse of the `(-k-1)`-th generator.

        EXAMPLES::

            sage: F = Gamma0(40).farey_symbol()
            sage: table = F.pairing_matrices_to_tietze_index()
            sage: table[12]
            (-2, -1)
            sage: F.pairing_matrices()[12]
            [  3  -1]
            [ 40 -13]
            sage: F.generators()[1]**-1
            [ -3   1]
            [-40  13]
        """
        gens_dict = {g: i+1 for i, g in enumerate(self.generators())}
        ans = []
        for pm in self.pairing_matrices():
            a, b, c, d = pm.matrix().list()
            newval = gens_dict.get(SL2Z([a, b, c, d]))
            if newval is not None:
                ans.append((newval, 1))
                continue
            newval = gens_dict.get(SL2Z([-a, -b, -c, -d]))
            if newval is not None:
                ans.append((newval, -1))
                continue
            newval = gens_dict.get(SL2Z([d, -b, -c, a]))
            if newval is not None:
                ans.append((-newval, 1))
                continue
            newval = gens_dict.get(SL2Z([-d, b, c, -a]))
            if newval is not None:
                ans.append((-newval, -1))
                continue
            raise RuntimeError("This should have not happened")
        return ans

    @cached_method
    def _get_minus_one(self):
        r"""
        If -I belongs to ``self``, return a Tietze word representing it.
        Otherwise return ``[]``.

        EXAMPLES::

            sage: Gamma1(30).farey_symbol()._get_minus_one()
            []
            sage: G = Gamma0(30).farey_symbol()
            sage: G._get_minus_one()
            [14]
            sage: g = G.generators()[13]
            sage: (-g.matrix()).is_one()
            True
            sage: G = Gamma(1).farey_symbol()
            sage: G._get_minus_one()
            [1, 1]
            sage: g = G.generators()[0]**2
            sage: (-g.matrix()).is_one()
            True
            sage: G = Gamma0(3).farey_symbol()
            sage: G._get_minus_one()
            [2, 2, 2]
            sage: g = G.generators()[1]**3
            sage: (-g.matrix()).is_one()
            True
        """
        for i, g in enumerate(self.generators()):
            m = g.matrix()
            if (-m).is_one():
                return [i + 1]
            t = m.trace()
            if t == 0:  # order = 4
                return 2 * [i + 1]
            elif t == 1:  # order = 6
                return 3 * [i + 1]
        return []

    def word_problem(self, M, output='standard', check=True):
        r"""
        Solve the word problem (up to sign) using this Farey symbol.

        INPUT:

        - ``M`` -- an element `M` of `\SL_2(\ZZ)`

        - ``output`` -- (default: ``'standard'``) should be one of
          ``'standard'``, ``'syllables'``, ``'gens'``.

        - ``check`` -- boolean (default: ``True``); whether to check for
          correct input and output

        OUTPUT:

        A solution to the word problem for the matrix `M`.
        The format depends on the ``output`` parameter, as follows.

        - ``'standard'`` returns the so called Tietze representation,
          which consists of a tuple of nonzero integers.  A positive
          integer `i` indicates the `i`-th generator (that is,
          ``self.generators()[i-1]``), while a negative integer `i`
          indicates the inverse of the `i`-th generator.
        - ``'syllables'`` returns a tuple of tuples of the form
          `(i, n)`, where `(i, n)` represents ``self.generators()[i] ^ n``,
          whose product equals `M` up to sign.
        - ``'gens'`` returns a tuple of pairs `(g, n)`, where `g` is a
          matrix and `n` an integer, such that the product of the
          matrices `g^n` equals `M` up to sign.

        EXAMPLES::

            sage: F = Gamma0(30).farey_symbol()
            sage: gens = F.generators()
            sage: g = gens[3] * gens[10] * gens[8]^-1 * gens[5]
            sage: g
            [-628597   73008]
            [-692130   80387]
            sage: F.word_problem(g)
            (4, 11, -9, 6)
            sage: g = gens[3] * gens[10]^2 * gens[8]^-1 * gens[5]
            sage: g
            [-5048053   586303]
            [-5558280   645563]
            sage: F.word_problem(g, output='gens')
            ((
            [109 -10]
            [120 -11], 1
            ),
             (
            [ 19  -7]
            [ 30 -11], 2
            ),
             (
            [ 49  -9]
            [ 60 -11], -1
            ),
             (
            [17 -2]
            [60 -7], 1
            ))
            sage: F.word_problem(g, output='syllables')
            ((3, 1), (10, 2), (8, -1), (5, 1))

        TESTS:

        Check that problem with forgotten generator is fixed::

            sage: from sage.misc.misc_c import prod
            sage: G = Gamma0(10)
            sage: F = G.farey_symbol()
            sage: g = G([-701,-137,4600,899])
            sage: g1 = prod(F.generators()[i]**a for i, a in F.word_problem(g, output='syllables'))
            sage: g == g1
            True

        Check that it works for GammaH as well (:issue:`19660`)::

            sage: G = GammaH(147, [8])
            sage: G.farey_symbol().word_problem(G([1,1,0,1]))
            (1,)

        Check that :issue:`20347` is solved::

            sage: # needs sage.groups
            sage: from sage.misc.misc_c import prod
            sage: G = ArithmeticSubgroup_Permutation(S2="(1,2)(3,4)", S3="(1,2,3)")
            sage: S = G.farey_symbol()
            sage: g1, g2 = S.generators()
            sage: g = g1^3 * g2^-2 * g1 * g2
            sage: S.word_problem(g)
            (2, 2, 2, 1, 1, 1, 2, 1, 2)
            sage: h = prod(S.generators()[i]**a for i, a in S.word_problem(g, output='syllables'))
            sage: g == h
            True

        """
        if output not in ['standard', 'syllables', 'gens']:
            raise ValueError('Unrecognized output format')
        if check:
            if M not in self.group:
                raise ValueError("Matrix ( %s ) is not in group ( %s )" % (M, self.group))
        cdef Integer a = M.d()
        cdef Integer b = -M.b()
        cdef Integer c = -M.c()
        cdef Integer d = M.a()
        cdef cpp_SL2Z *cpp_beta = new cpp_SL2Z(1, 0, 0, 1)
        sig_on()
        result = self.this_ptr.word_problem(a.value, b.value, c.value, d.value, cpp_beta)
        sig_off()
        beta = convert_to_SL2Z(cpp_beta[0])**-1
        mbeta = SL2Z([-beta.a(), -beta.b(), -beta.c(), -beta.d()])
        V = self.pairing_matrices_to_tietze_index()
        sgn = 1
        tietze = []
        for o in result:
            if o > 0:
                tietze.append(V[o-1][0])
                sgn *= V[o-1][1]
            else:
                tietze.append(-V[-o-1][0])
                sgn *= V[-o-1][1]
        if sgn == -1:
            beta, mbeta = mbeta, beta

        gens_dict = {g: i+1 for i, g in enumerate(self.generators())}
        extra_tietze = []
        if beta.is_one():
            found = True
        elif mbeta.is_one():
            found = True
            extra_tietze = self._get_minus_one()
        else:
            found = False
        if not found:
            newval = gens_dict.get(beta)
            if newval is not None:
                found = True
                extra_tietze = [newval]
        if not found:
            newval = gens_dict.get(beta**-1)
            if newval is not None:
                found = True
                extra_tietze = [-newval]
        if not found:
            newval = gens_dict.get(mbeta)
            if newval is not None:
                found = True
                extra_tietze = [newval] + self._get_minus_one()
        if not found:
            newval = gens_dict.get(mbeta**-1)
            if newval is not None:
                found = True
                extra_tietze = [-newval] + self._get_minus_one()
        tietze.extend(extra_tietze)
        tietze.reverse()
        gens = self.generators()
        if check:
            tmp = SL2Z([1, 0, 0, 1])
            for i in range(len(tietze)):
                t = tietze[i]
                tmp = tmp * gens[t-1] if t > 0 else tmp * gens[-t-1]**-1
            assert tmp.matrix() == M.matrix(), '%s %s %s' % (tietze, tmp.matrix(), M.matrix())
        if output == 'standard':
            return tuple(tietze)
        if output == 'syllables':
            return tuple((a-1, len(list(g))) if a > 0 else (-a-1, -len(list(g))) for a, g in groupby(tietze))
        else:  # output == 'gens'
            return tuple((gens[a-1], len(list(g))) if a > 0 else (gens[-a-1], -len(list(g))) for a, g in groupby(tietze))

    def __contains__(self, M) -> bool:
        r"""
        Test if element is in the arithmetic group of the Farey symbol
        via LLT algorithm.

        EXAMPLES::

            sage: SL2Z([0, -1, 1, 0]) in FareySymbol(Gamma0(6))
            False

            sage: SL2Z([1, 1, 0, 1]) in FareySymbol(Gamma0(6))
            True
        """
        cdef Integer a = M.a()
        cdef Integer b = M.b()
        cdef Integer c = M.c()
        cdef Integer d = M.d()
        sig_on()
        result = self.this_ptr.is_element(a.value, b.value, c.value, d.value)
        sig_off()
        return result

    def __richcmp__(self, other, op) -> bool:
        r"""
        Compare ``self`` to ``other``.

        EXAMPLES::

            sage: FareySymbol(Gamma(2)) == FareySymbol(Gamma0(7))
            False

            sage: FareySymbol(Gamma0(23)) == loads(dumps(FareySymbol(Gamma0(23))))
            True
        """
        if not isinstance(other, Farey):
            return NotImplemented

        cosetA = self.coset_reps()
        cosetB = other.coset_reps()
        if cosetA != cosetB:
            return richcmp_not_equal(cosetA, cosetB, op)

        cuspA = self.cusps()
        cuspB = other.cusps()
        if cuspA != cuspB:
            return richcmp_not_equal(cuspA, cuspB, op)

        return PyObject_RichCompare(self.fractions(), other.fractions(), op)

    def __reduce__(self):
        r"""
        Serialization for pickling::

            sage: FareySymbol(Gamma0(4)).__reduce__()
            (<class 'sage.modular.arithgroup.farey_symbol.Farey'>, ...))
        """
        return Farey, (self.group, self.this_ptr.dumps())

    def __repr__(self) -> str:
        r"""
        Return the string representation of ``self``.

        EXAMPLES::

            sage: FareySymbol(Gamma0(23)).__repr__()
            'FareySymbol(Congruence Subgroup Gamma0(23))'
        """
        if hasattr(self.group, "_repr_"):
            return "FareySymbol(%s)" % self.group._repr_()
        elif hasattr(self.group, "__repr__"):
            return "FareySymbol(%r)" % self.group
        else:
            return "FareySymbol(?)"

    def _latex_(self, forced_format=None):
        r"""
        Return the LaTeX representation of ``self``.

        INPUT:

        - ``forced_format`` -- a format string ('plain' or 'xymatrix')
          or ``None``

        EXAMPLES::

            sage: FareySymbol(Gamma0(11))._latex_(forced_format='plain')
            '\\left( -\\infty\\underbrace{\\quad}_{1} 0\\underbrace{\\quad}_{2} \\frac{1}{3}\\underbrace{\\quad}_{3} \\frac{1}{2}\\underbrace{\\quad}_{2} \\frac{2}{3}\\underbrace{\\quad}_{3} 1\\underbrace{\\quad}_{1} \\infty\\right)'
            sage: FareySymbol(Gamma0(11))._latex_(forced_format='xymatrix')
            '\\begin{xy}\\xymatrix{& -\\infty \\ar@{-}@/_1pc/[r]_{1}& 0 \\ar@{-}@/_1pc/[r]_{2}& \\frac{1}{3} \\ar@{-}@/_1pc/[r]_{3}& \\frac{1}{2} \\ar@{-}@/_1pc/[r]_{2}& \\frac{2}{3} \\ar@{-}@/_1pc/[r]_{3}& 1 \\ar@{-}@/_1pc/[r]_{1}& \\infty }\\end{xy}'

            sage: 'xymatrix' in FareySymbol(Gamma0(11))._latex_()
            True
        """
        if forced_format == 'plain':
            # output not using xymatrix
            s = r'\left( -\infty'
            a = [x._latex_() for x in self.fractions()] + [r'\infty']
            b = self.pairings()
            for i in range(len(a)):
                u = b[i]
                if u == -3:
                    u = r'\bullet'
                elif u == -2:
                    u = r'\circ'
                s += r'\underbrace{\quad}_{%s} %s' % (u, a[i])
            return s + r'\right)'
        else:
            # output using xymatrix
            s = r'\begin{xy}\xymatrix{& -\infty '
            f = [x._latex_() for x in self.fractions()] + [r'\infty']
            f.reverse()
            for p in self.pairings():
                if p >= 0:
                    s += r'\ar@{-}@/_1pc/[r]_{%s}' % p
                elif p == -2:
                    s += r'\ar@{-}@/_1pc/[r]_{\circ}'
                elif p == -3:
                    s += r'\ar@{-}@/_1pc/[r]_{\bullet}'
                s += r'& %s ' % f.pop()
            s += r'}\end{xy}'
            return s

    def index(self):
        r"""
        Return the index of the arithmetic group of the FareySymbol
        in `\PSL_2(\ZZ)`.

        EXAMPLES::

            sage: [FareySymbol(Gamma0(n)).index() for n in range(1, 16)]
            [1, 3, 4, 6, 6, 12, 8, 12, 12, 18, 12, 24, 14, 24, 24]
        """
        return self.this_ptr.index()

    def genus(self):
        r"""
        Return the genus of the arithmetic group of the FareySymbol.

        EXAMPLES::

            sage: [FareySymbol(Gamma0(n)).genus() for n in range(16, 32)]
            [0, 1, 0, 1, 1, 1, 2, 2, 1, 0, 2, 1, 2, 2, 3, 2]
        """
        return self.this_ptr.genus()

    def level(self):
        r"""
        Return the level of the arithmetic group of the FareySymbol.

        EXAMPLES::

            sage: [FareySymbol(Gamma0(n)).level() for n in range(1, 16)]
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
        """
        return self.this_ptr.level()

    def nu2(self):
        r"""
        Return the number of elliptic points of order two.

        EXAMPLES::

            sage: [FareySymbol(Gamma0(n)).nu2() for n in range(1, 16)]
            [1, 1, 0, 0, 2, 0, 0, 0, 0, 2, 0, 0, 2, 0, 0]
        """
        return self.this_ptr.nu2()

    def nu3(self):
        r"""
        Return the number of elliptic points of order three.

        EXAMPLES::

            sage: [FareySymbol(Gamma0(n)).nu3() for n in range(1, 16)]
            [1, 0, 1, 0, 0, 0, 2, 0, 0, 0, 0, 0, 2, 0, 0]
        """
        return self.this_ptr.nu3()

    def coset_reps(self):
        r"""
        Left coset of the arithmetic group of the FareySymbol.

        EXAMPLES:

        Calculate the left coset of `\Gamma_0(6)`::

            sage: FareySymbol(Gamma0(6)).coset_reps()
            [
            [1 0]  [ 3 -1]  [ 2 -1]  [ 1 -1]  [ 2 -1]  [ 3 -2]  [ 1 -1]  [-1  0]
            [0 1], [ 1  0], [ 1  0], [ 1  0], [ 3 -1], [ 2 -1], [ 2 -1], [ 2 -1],
            [ 1 -1]  [ 0 -1]  [-1  0]  [-2  1]
            [ 3 -2], [ 1 -1], [ 1 -1], [ 1 -1]
            ]
        """
        return self.this_ptr.get_coset()

    def generators(self):
        r"""
        Minimal set of generators of the group of the FareySymbol.

        EXAMPLES:

        Calculate the generators of `\Gamma_0(6)`::

            sage: FareySymbol(Gamma0(6)).generators()
            [
            [1 1]  [ 5 -1]  [ 7 -3]  [-1  0]
            [0 1], [ 6 -1], [12 -5], [ 0 -1]
            ]

        Calculate the generators of `\SL_2(\ZZ)`::

            sage: FareySymbol(SL2Z).generators()
            [
            [ 0 -1]  [ 0 -1]
            [ 1  0], [ 1 -1]
            ]

        The unique index 2 even subgroup and index 4 odd subgroup each get handled correctly::

            sage: # needs sage.groups
            sage: FareySymbol(ArithmeticSubgroup_Permutation(S2="(1,2)", S3="()")).generators()
            [
            [ 0  1]  [-1  1]
            [-1 -1], [-1  0]
            ]
            sage: FareySymbol(ArithmeticSubgroup_Permutation(S2="(1,2, 3, 4)", S3="(1,3)(2,4)")).generators()
            [
            [ 0  1]  [-1  1]
            [-1 -1], [-1  0]
            ]
        """
        return self.this_ptr.get_generators()

    def fractions(self):
        r"""
        Fractions of the FareySymbol.

        EXAMPLES::

            sage: FareySymbol(Gamma(4)).fractions()
            [0, 1/2, 1, 3/2, 2, 5/2, 3, 7/2, 4]
        """
        return self.this_ptr.get_fractions()

    def pairings(self):
        r"""
        Pairings of the sides of the fundamental domain of the Farey symbol
        of the arithmetic group.

        The sides of the hyperbolic polygon are
        numbered 0, 1, ... from left to right. Conventions: even pairings are
        denoted by -2, odd pairings by -3 while free pairings are denoted by
        an integer number greater than zero.

        EXAMPLES:

        Odd pairings::

            sage: FareySymbol(Gamma0(7)).pairings()
            [1, -3, -3, 1]

        Even and odd pairings::

            FareySymbol(Gamma0(13)).pairings()
            [1, -3, -2, -2, -3, 1]

        Only free pairings::

            sage: FareySymbol(Gamma0(23)).pairings()
            [1, 2, 3, 5, 3, 4, 2, 4, 5, 1]
        """
        return self.this_ptr.get_pairings()

    def paired_sides(self):
        r"""
        Pairs of index of the sides of the fundamental domain of the
        Farey symbol of the arithmetic group. The sides of the
        hyperbolic polygon are numbered 0, 1, ... from left to right.

        .. image:: ../../../media/pairing.png

        EXAMPLES::

            sage: FareySymbol(Gamma0(11)).paired_sides()
            [(0, 5), (1, 3), (2, 4)]

        indicating that the side 0 is paired with 5, 1 with 3 and 2 with 4.
        """
        return self.this_ptr.get_paired_sides()

    def pairing_matrices(self):
        r"""
        Pairing matrices of the sides of the fundamental domain. The sides
        of the hyperbolic polygon are numbered 0, 1, ... from left to right.

        EXAMPLES::

            sage: FareySymbol(Gamma0(6)).pairing_matrices()
            [
            [1 1]  [ 5 -1]  [ 7 -3]  [ 5 -3]  [ 1 -1]  [-1  1]
            [0 1], [ 6 -1], [12 -5], [12 -7], [ 6 -5], [ 0 -1]
            ]
        """

        return self.this_ptr.get_pairing_matrices()

    def cusps(self):
        r"""
        Cusps of the FareySymbol.

        EXAMPLES::

            sage: FareySymbol(Gamma0(6)).cusps()
            [0, 1/3, 1/2, Infinity]
        """
        return self.this_ptr.get_cusps()+[Cusp(infinity)]

    def cusp_widths(self):
        r"""
        Cusps widths of the FareySymbol.

        EXAMPLES::

            sage: FareySymbol(Gamma0(6)).cusp_widths()
            [6, 2, 3, 1]
        """
        return self.this_ptr.get_cusp_widths()

    def cusp_class(self, c):
        r"""
        Cusp class of a cusp in the FareySymbol.

        INPUT:

        - ``c`` -- a cusp

        EXAMPLES::

            sage: FareySymbol(Gamma0(12)).cusp_class(Cusp(1, 12))
            5
        """
        cusp = Cusp(c)
        cdef Integer p = cusp.numerator()
        cdef Integer q = cusp.denominator()
        sig_on()
        result = self.this_ptr.get_cusp_class(p.value, q.value)
        sig_off()
        return result

    def reduce_to_cusp(self, r):
        r"""
        Transformation of a rational number to cusp representative.

        INPUT:

        - ``r`` -- a rational number

        EXAMPLES::

            sage: FareySymbol(Gamma0(12)).reduce_to_cusp(5/8)
            [ 5  -3]
            [12  -7]

        Reduce 11/17 to a cusp of for HsuExample10()::

            sage: # needs sage.groups
            sage: from sage.modular.arithgroup.arithgroup_perm import HsuExample10
            sage: f = FareySymbol(HsuExample10())
            sage: f.reduce_to_cusp(11/17)
            [14 -9]
            [-3  2]
            sage: _.acton(11/17)
            1
            sage: f.cusps()[f.cusp_class(11/17)]
            1
        """
        cdef Integer p = r.numerator()
        cdef Integer q = r.denominator()
        sig_on()
        result = self.this_ptr.get_transformation_to_cusp(p.value, q.value)
        sig_off()
        return result

    @options(alpha=1, fill=True, thickness=1, color='lightgray',
             color_even='white',
             zorder=2, linestyle='solid', show_pairing=True,
             tesselation='Dedekind', ymax=1)
    def fundamental_domain(self, **options):
        r"""
        Plot a fundamental domain of an arithmetic subgroup of
        `\PSL_2(\ZZ)` corresponding to the Farey symbol.

        OPTIONS:

        - ``fill`` -- boolean (default: ``True``); fill the fundamental domain

        - ``linestyle`` -- string (default: ``'solid'``); the style of the line,
          which is one of 'dashed', 'dotted', 'solid', 'dashdot', or '--',
          ':', '-', '-.', respectively

        - ``color`` -- (default: ``'lightgray'``) fill color for odd part of
          Dedekind tesselation

        - ``show_pairing`` -- boolean (default: ``True``); flag for pairing

        - ``tesselation`` -- (default: ``'Dedekind'``) the type of
          hyperbolic tesselation which is one of
          ``'coset'``, ``'Dedekind'`` or ``None`` respectively

        - ``color_even`` -- fill color for even parts of Dedekind
          tesselation (default: ``'white'``); ignored for other tesselations

        - ``thickness`` -- float (default: `1`) the thickness of the line

        - ``ymax`` -- float (default: `1`) maximal height

        EXAMPLES:

        For example, to plot the fundamental domain of `\Gamma_0(11)`
        with pairings use the following command::

            sage: FareySymbol(Gamma0(11)).fundamental_domain()                          # needs sage.plot sage.symbolic
            Graphics object consisting of 54 graphics primitives

        indicating that side 1 is paired with side 3 and side 2 is
        paired with side 4, see also :meth:`.paired_sides`.

        To plot the fundamental domain of `\Gamma(3)` without pairings
        use the following command::

            sage: FareySymbol(Gamma(3)).fundamental_domain(show_pairing=False)          # needs sage.plot sage.symbolic
            Graphics object consisting of 48 graphics primitives

        Plot the fundamental domain of `\Gamma_0(23)` showing the left
        coset representatives::

            sage: FareySymbol(Gamma0(23)).fundamental_domain(tesselation='coset')       # needs sage.plot sage.symbolic
            Graphics object consisting of 58 graphics primitives

        The same as above but with a custom linestyle::

            sage: FareySymbol(Gamma0(23)).fundamental_domain(tesselation='coset',       # needs sage.plot sage.symbolic
            ....:                                            linestyle=':',
            ....:                                            thickness='2')
            Graphics object consisting of 58 graphics primitives
        """
        from sage.plot.graphics import Graphics
        from sage.plot.colors import rainbow
        from sage.plot.hyperbolic_arc import hyperbolic_arc
        from sage.plot.hyperbolic_polygon import hyperbolic_triangle

        I = CC(0, 1)
        w = RR(3).sqrt()
        L = 1000
        g = Graphics()
        # show coset
        for x in self.coset_reps():
            a, b, c, d = x[1, 1], -x[0, 1], -x[1, 0], x[0, 0]
            A, B = CC(0, L), CC(0, L)
            if d != 0:
                A = b / d
            if c != 0:
                B = a / c
            C = (a*c+b*d+(a*d+b*c)/2+I*w/2)/(c*c+c*d+d*d)
            D = (a*c+b*d + CC(0, 1))/(c*c+d*d)
            if options['tesselation'] == 'Dedekind':
                g += hyperbolic_triangle(A, D, C,
                                         alpha=options['alpha'],
                                         color=options['color'],
                                         fill=options['fill'],
                                         linestyle=options['linestyle'],
                                         thickness=options['thickness'])
                g += hyperbolic_triangle(D, C, B,
                                         alpha=options['alpha'],
                                         color=options['color_even'],
                                         fill=options['fill'],
                                         linestyle=options['linestyle'],
                                         thickness=options['thickness'])
                g += hyperbolic_triangle(A, D, C, color='gray')
                g += hyperbolic_triangle(D, C, B, color='gray')
            elif options['tesselation'] == 'coset':
                g += hyperbolic_triangle(A, B, C,
                                         alpha=options['alpha'],
                                         color=options['color'],
                                         fill=options['fill'],
                                         linestyle=options['linestyle'],
                                         thickness=options['thickness'])
                g += hyperbolic_triangle(A, B, C, color='gray',
                                         linestyle=options['linestyle'],
                                         thickness=options['thickness'])
            else:
                g += hyperbolic_triangle(A, B, C,
                                         alpha=options['alpha'],
                                         color=options['color'],
                                         fill=options['fill'],
                                         linestyle=options['linestyle'],
                                         thickness=options['thickness'])
        # show pairings
        p = self.pairings()
        x = self.fractions()
        if options['show_pairing']:
            rc = rainbow(max(p)-min(p)+1)
            if p[0] > 0:
                g += hyperbolic_arc(CC(0, L), x[0], color=rc[p[0]-min(p)],
                                    linestyle=options['linestyle'],
                                    thickness=options['thickness'])
            if p[-1] > 0:
                g += hyperbolic_arc(CC(0, L), x[-1], color=rc[p[-1]-min(p)],
                                    linestyle=options['linestyle'],
                                    thickness=options['thickness'])
            for i in range(len(x)-1):
                if p[i+1] > 0:
                    g += hyperbolic_arc(x[i], x[i+1], color=rc[p[i+1]-min(p)],
                                        linestyle=options['linestyle'],
                                        thickness=options['thickness'])
        d = g.get_minmax_data()
        g.set_axes_range(d['xmin'], d['xmax'], 0, options['ymax'])
        g.set_aspect_ratio(1)
        return g


# ----- conversions ---------------------------------

cdef public long convert_to_long(n) noexcept:
    cdef long m = n
    return m

cdef public object convert_to_Integer(mpz_class a):
    A = Integer()
    A.set_from_mpz(a.get_mpz_t())
    return A

cdef public object convert_to_rational(mpq_class r):
    a = Integer()
    a.set_from_mpz(r.get_num_mpz_t())
    b = Integer()
    b.set_from_mpz(r.get_den_mpz_t())
    return a/b

cdef public object convert_to_cusp(mpq_class r):
    a = Integer()
    a.set_from_mpz(r.get_num_mpz_t())
    b = Integer()
    b.set_from_mpz(r.get_den_mpz_t())
    return Cusp(a/b)

cdef public object convert_to_SL2Z(cpp_SL2Z M):
    a = convert_to_Integer(M.a())
    b = convert_to_Integer(M.b())
    c = convert_to_Integer(M.c())
    d = convert_to_Integer(M.d())
    return SL2Z([a, b, c, d])



# sage_setup: distribution = sagemath-modules
r"""
Free module morphisms

The class :class:`FiniteRankFreeModuleMorphism` implements homomorphisms
between two free modules of finite rank over the same commutative ring.

The subclass :class:`FiniteRankFreeModuleEndomorphism` implements the
special case of endomorphisms.

AUTHORS:

- Eric Gourgoulhon, Michal Bejger (2014-2015): initial version
- Matthias Koeppe (2024): add subclass :class:`FiniteRankFreeModuleEndomorphism`

REFERENCES:

- Chap. 13, 14 of R. Godement : *Algebra* [God1968]_
- Chap. 3 of S. Lang : *Algebra* [Lan2002]_
"""
# *****************************************************************************
#       Copyright (C) 2015 Eric Gourgoulhon <eric.gourgoulhon@obspm.fr>
#       Copyright (C) 2015 Michal Bejger <bejger@camk.edu.pl>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *****************************************************************************
from __future__ import annotations

from typing import TYPE_CHECKING

from sage.categories.morphism import Morphism
from sage.misc.lazy_attribute import lazy_attribute
from sage.rings.integer import Integer

if TYPE_CHECKING:
    from sage.tensor.modules.free_module_element import FiniteRankFreeModuleElement


class FiniteRankFreeModuleMorphism(Morphism):
    r"""
    Homomorphism between free modules of finite rank over a commutative ring.

    An instance of this class is a homomorphism

    .. MATH::

        \phi:\ M \longrightarrow N,

    where `M` and `N` are two free modules of finite rank over the same
    commutative ring `R`.

    This is a Sage *element* class, the corresponding *parent* class being
    :class:`~sage.tensor.modules.free_module_homset.FreeModuleHomset`.

    For the special case of endomorphisms (`M=N`), use the subclass
    :class:`FiniteRankFreeModuleEndomorphism`.

    INPUT:

    - ``parent`` -- Hom-set Hom(M,N) to which the homomorphism belongs
    - ``matrix_rep`` -- matrix representation of the homomorphism with
      respect to the bases ``bases``; this entry can actually
      be any material from which a matrix of size rank(N)*rank(M) of
      elements of `R` can be constructed; the *columns* of the matrix give
      the images of the basis of `M` (see the convention in the example below)
    - ``bases`` -- (default: ``None``) pair (basis_M, basis_N) defining the
      matrix representation, basis_M being a basis of module `M` and
      basis_N a basis of module `N` ; if None the pair formed by the
      default bases of each module is assumed.
    - ``name`` -- (default: ``None``) string; name given to the homomorphism
    - ``latex_name`` -- (default: ``None``) string; LaTeX symbol to denote the
      homomorphism. If ``None``, ``name`` will be used.

    EXAMPLES:

    A homomorphism between two free modules over `\ZZ` is constructed
    as an element of the corresponding hom-set, by means of the function
    ``__call__``::

        sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
        sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
        sage: e = M.basis('e') ; f = N.basis('f')
        sage: H = Hom(M,N) ; H
        Set of Morphisms from Rank-3 free module M over the Integer Ring
         to Rank-2 free module N over the Integer Ring
         in Category of finite dimensional modules over Integer Ring
        sage: phi = H([[2,-1,3], [1,0,-4]], name='phi', latex_name=r'\phi') ; phi
        Generic morphism:
          From: Rank-3 free module M over the Integer Ring
          To:   Rank-2 free module N over the Integer Ring

    Since no bases have been specified in the argument list, the provided
    matrix is relative to the default bases of modules M and N, so that
    the above is equivalent to::

        sage: phi = H([[2,-1,3], [1,0,-4]], bases=(e,f), name='phi',
        ....:         latex_name=r'\phi') ; phi
        Generic morphism:
          From: Rank-3 free module M over the Integer Ring
          To:   Rank-2 free module N over the Integer Ring

    An alternative way to construct a homomorphism is to call the method
    :meth:`~sage.tensor.modules.finite_rank_free_module.FiniteRankFreeModule.hom`
    on the domain::

        sage: phi = M.hom(N, [[2,-1,3], [1,0,-4]], bases=(e,f), name='phi',
        ....:             latex_name=r'\phi') ; phi
        Generic morphism:
          From: Rank-3 free module M over the Integer Ring
          To:   Rank-2 free module N over the Integer Ring

    The parent of a homomorphism is of course the corresponding hom-set::

        sage: phi.parent() is H
        True
        sage: phi.parent() is Hom(M,N)
        True

    Due to Sage's category scheme, the actual class of the homomorphism ``phi``
    is a derived class of :class:`FiniteRankFreeModuleMorphism`::

        sage: type(phi)
        <class 'sage.tensor.modules.free_module_homset.FreeModuleHomset_with_category_with_equality_by_id.element_class'>
        sage: isinstance(phi, sage.tensor.modules.free_module_morphism.FiniteRankFreeModuleMorphism)
        True

    The domain and codomain of the homomorphism are returned respectively by
    the methods ``domain()`` and ``codomain()``, which are implemented as
    Sage's constant functions::

        sage: phi.domain()
        Rank-3 free module M over the Integer Ring
        sage: phi.codomain()
        Rank-2 free module N over the Integer Ring
        sage: type(phi.domain)
        <... 'sage.misc.constant_function.ConstantFunction'>

    The matrix of the homomorphism with respect to a pair of bases is
    returned by the method :meth:`matrix`::

        sage: phi.matrix(e,f)
        [ 2 -1  3]
        [ 1  0 -4]

    The convention is that the columns of this matrix give the components of
    the images of the elements of basis ``e`` w.r.t basis ``f``::

        sage: phi(e[0]).display()
        phi(e_0) = 2 f_0 + f_1
        sage: phi(e[1]).display()
        phi(e_1) = -f_0
        sage: phi(e[2]).display()
        phi(e_2) = 3 f_0 - 4 f_1

    Test of the module homomorphism laws::

        sage: phi(M.zero()) == N.zero()
        True
        sage: u = M([1,2,3], basis=e, name='u') ; u.display()
        u = e_0 + 2 e_1 + 3 e_2
        sage: v = M([-2,1,4], basis=e, name='v') ; v.display()
        v = -2 e_0 + e_1 + 4 e_2
        sage: phi(u).display()
        phi(u) = 9 f_0 - 11 f_1
        sage: phi(v).display()
        phi(v) = 7 f_0 - 18 f_1
        sage: phi(3*u + v).display()
        34 f_0 - 51 f_1
        sage: phi(3*u + v) == 3*phi(u) + phi(v)
        True
    """
    def __init__(self, parent, matrix_rep, bases=None, name=None,
                 latex_name=None, is_identity=False):
        r"""
        TESTS:

        Generic homomorphism::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: from sage.tensor.modules.free_module_morphism import FiniteRankFreeModuleMorphism
            sage: phi = FiniteRankFreeModuleMorphism(Hom(M,N), [[1,0,-3], [2,1,4]],
            ....:                                    name='phi', latex_name=r'\phi')
            sage: phi
            Generic morphism:
              From: Rank-3 free module M over the Integer Ring
              To:   Rank-2 free module N over the Integer Ring
            sage: phi.matrix(e,f)
            [ 1  0 -3]
            [ 2  1  4]
            sage: latex(phi)
            \phi
        """
        if is_identity:
            raise TypeError('use the subclass FiniteRankFreeModuleEndomorphism for the identity morphis')
        from sage.matrix.constructor import matrix
        from sage.misc.constant_function import ConstantFunction
        Morphism.__init__(self, parent)
        fmodule1 = parent.domain()
        fmodule2 = parent.codomain()
        if bases is None:
            def_basis1 = fmodule1.default_basis()
            if def_basis1 is None:
                raise ValueError("the {} has no default ".format(fmodule1) +
                                 "basis")
            def_basis2 = fmodule2.default_basis()
            if def_basis2 is None:
                raise ValueError("the {} has no default ".format(fmodule2) +
                                 "basis")
            bases = (def_basis1, def_basis2)
        else:
            bases = tuple(bases)  # insures bases is a tuple
            if len(bases) != 2:
                raise TypeError("the argument bases must contain 2 bases")
            if bases[0] not in fmodule1.bases():
                raise TypeError("{} is not a basis on the {}".format(bases[0],
                                                                     fmodule1))
            if bases[1] not in fmodule2.bases():
                raise TypeError("{} is not a basis on the {}".format(bases[1],
                                                                     fmodule2))
        ring = parent.base_ring()
        n1 = fmodule1.rank()
        n2 = fmodule2.rank()
        # Construction of a generic morphism
        self._is_identity = False
        if isinstance(matrix_rep, ConstantFunction):
            # the zero morphism
            if matrix_rep().is_zero():
                matrix_rep = 0
        self._matrices = {bases: matrix(ring, n2, n1, matrix_rep)}
        self._name = name
        if latex_name is None:
            self._latex_name = self._name
        else:
            self._latex_name = latex_name

    #
    # SageObject methods
    #

    def _latex_(self):
        r"""
        LaTeX representation of the object.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]], name='phi',
            ....:             latex_name=r'\Phi')
            sage: phi._latex_()
            '\\Phi'
            sage: latex(phi)  # indirect doctest
            \Phi

        ::

            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]], name='F')
            sage: phi._latex_()
            'F'

        ::

            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]])
            sage: phi._latex_()
            '\\mbox{Generic morphism:\n  From: Rank-3 free module M over the Integer Ring\n  To:   Rank-2 free module N over the Integer Ring}'
        """
        if self._latex_name is None:
            return r'\mbox{' + str(self) + r'}'
        else:
            return self._latex_name

    def __eq__(self, other):
        r"""
        Comparison (equality) operator.

        INPUT:

        - ``other`` -- a free module morphism (or 0)

        OUTPUT: ``True`` if ``self`` is equal to ``other`` and ``False`` otherwise

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]], name='phi',
            ....:             latex_name=r'\phi')
            sage: psi = M.hom(N, [[-1,2,0], [5,1,2]])
            sage: phi.__eq__(psi)
            True
            sage: phi == psi
            True
            sage: phi.__eq__(phi)
            True
            sage: phi.__eq__(+phi)
            True
            sage: psi = M.hom(N, [[1,1,0], [4,1,3]])
            sage: phi.__eq__(psi)
            False
            sage: phi.__eq__(-phi)
            False

        Comparison of homomorphisms defined on different bases::

            sage: a = M.automorphism() ; a[0,2], a[1,0], a[2,1] = 1, -1, -1
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: psi = M.hom(N, [[-2,0,-1], [-1,-2, 5]], bases=(ep,f))
            sage: phi.__eq__(psi)
            True
            sage: phi.matrix(e,f) == psi.matrix(e,f)  # check
            True

        Comparison of homomorphisms having the same matrix but defined on
        different modules::

            sage: N1 = FiniteRankFreeModule(ZZ, 2, name='N1')
            sage: f1 = N1.basis('f')
            sage: phi1 = M.hom(N1, [[-1,2,0], [5,1,2]])
            sage: phi.matrix() == phi1.matrix() # same matrix in the default bases
            True
            sage: phi.__eq__(phi1)
            False

        Comparison to zero::

            sage: phi.__eq__(0)
            False
            sage: phi = M.hom(N, 0)
            sage: phi.__eq__(0)
            True
            sage: phi == 0
            True
            sage: phi.__eq__(Hom(M,N).zero())
            True
        """
        if isinstance(other, (int, Integer)): # other should be 0
            if other == 0:
                return self.is_zero()
            else:
                return False
        elif not isinstance(other, FiniteRankFreeModuleMorphism):
            return False
        elif self.parent() != other.parent():
            return False
        else:
            bases = self._common_bases(other)
            if bases is None:
                raise ValueError("no common pair of bases has been found to " +
                                 "compare {} and {}".format(self, other))
            return bool( self.matrix(*bases) == other.matrix(*bases) )

    def __ne__(self, other):
        r"""
        Inequality operator.

        INPUT:

        - ``other`` -- a free module morphism (or 0)

        OUTPUT:

        - ``True`` if ``self`` is different from ``other`` and ``False``
          otherwise

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]], name='phi',
            ....:             latex_name=r'\phi')
            sage: psi = M.hom(N, [[-1,2,0], [5,1,2]])
            sage: phi.__ne__(psi)
            False
            sage: psi = M.hom(N, [[1,1,0], [4,1,3]])
            sage: phi.__ne__(psi)
            True
            sage: phi != psi
            True
            sage: phi.__ne__('junk')
            True
            sage: Hom(M,N).zero().__ne__(0)
            False
        """
        return not self == other

    #
    # Required module methods
    #

    def __bool__(self):
        r"""
        Return ``True`` if ``self`` is nonzero and ``False`` otherwise.

        This method is called by self.is_zero().

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: phi = M.hom(N, [[2,-1,3], [1,0,-4]])
            sage: bool(phi)
            True
            sage: phi.is_zero() # indirect doctest
            False
            sage: phi = M.hom(N, 0)
            sage: bool(phi)
            False
            sage: phi.is_zero() # indirect doctest
            True
            sage: bool(Hom(M,N).zero())
            False
        """
        # Some matrix representation is picked at random:
        matrix_rep = next(iter(self._matrices.values()))
        return not matrix_rep.is_zero()

    def _add_(self, other):
        r"""
        Homomorphism addition.

        INPUT:

        - ``other`` -- a free module morphism (same parent as ``self``)

        OUTPUT: the homomorphism resulting from the addition of ``self`` and ``other``

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]], name='phi',
            ....:             latex_name=r'\phi')
            sage: psi = M.hom(N, [[1,1,0], [4,1,3]])
            sage: s = phi._add_(psi) ; s
            Generic morphism:
              From: Rank-3 free module M over the Integer Ring
              To:   Rank-2 free module N over the Integer Ring
            sage: s.matrix(e,f)
            [0 3 0]
            [9 2 5]
            sage: s.matrix(e,f) == phi.matrix(e,f) + psi.matrix(e,f)  # check
            True
            sage: s == phi + psi  # indirect doctest
            True

        Addition of homomorphisms defined on different bases::

            sage: a = M.automorphism() ; a[0,2], a[1,0], a[2,1] = 1, -1, -1
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: b = N.automorphism() ; b[0,1], b[1,0] = -1, 1
            sage: fp = f.new_basis(b, 'fp', latex_symbol="f'")
            sage: psi = M.hom(N, [[-2,0,-1], [-1,-2, 5]], bases=(ep,fp))
            sage: s = phi._add_(psi) ; s
            Generic morphism:
              From: Rank-3 free module M over the Integer Ring
              To:   Rank-2 free module N over the Integer Ring
            sage: s.matrix(e,f)
            [-6  1 -2]
            [ 4  3  2]
            sage: s.matrix(e,f) == phi.matrix(e,f) + psi.matrix(e,f)  # check
            True
            sage: s == phi + psi  # indirect doctest
            True

        Other tests::

            sage: phi._add_(Hom(M,N).zero()) == phi
            True
        """
        # No need for consistency checks since self and other are guaranteed
        # to have the same parents
        bases = self._common_bases(other)
        if bases is None:
            raise ValueError("no common pair of bases has been found to " +
                             "add {} and {}".format(self, other))
        # Addition at the matrix level:
        resu_mat = self._matrices[bases] + other._matrices[bases]
        if self._name is not None and other._name is not None:
            resu_name = self._name + '+' + other._name
        else:
            resu_name = None
        if self._latex_name is not None and other._latex_name is not None:
            resu_latex_name = self._latex_name + '+' + other._latex_name
        else:
            resu_latex_name = None
        return self.__class__(self.parent(), resu_mat, bases=bases,
                              name=resu_name, latex_name=resu_latex_name)

    def _sub_(self, other):
        r"""
        Homomorphism subtraction.

        INPUT:

        - ``other`` -- a free module morphism (same parent as ``self``)

        OUTPUT:

        - the homomorphism resulting from the subtraction of ``other`` from
          ``self``

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]], name='phi',
            ....:             latex_name=r'\phi')
            sage: psi = M.hom(N, [[1,1,0], [4,1,3]])
            sage: s = phi._sub_(psi) ; s
            Generic morphism:
              From: Rank-3 free module M over the Integer Ring
              To:   Rank-2 free module N over the Integer Ring
            sage: s.matrix(e,f)
            [-2  1  0]
            [ 1  0 -1]
            sage: s.matrix(e,f) == phi.matrix(e,f) - psi.matrix(e,f)  # check
            True
            sage: s == phi - psi  # indirect doctest
            True

        Subtraction of homomorphisms defined on different bases::

            sage: a = M.automorphism() ; a[0,2], a[1,0], a[2,1] = 1, -1, -1
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: b = N.automorphism() ; b[0,1], b[1,0] = -1, 1
            sage: fp = f.new_basis(b, 'fp', latex_symbol="f'")
            sage: psi = M.hom(N, [[-2,0,-1], [-1,-2, 5]], bases=(ep,fp))
            sage: s = phi._sub_(psi) ; s
            Generic morphism:
              From: Rank-3 free module M over the Integer Ring
              To:   Rank-2 free module N over the Integer Ring
            sage: s.matrix(e,f)
            [ 4  3  2]
            [ 6 -1  2]
            sage: s.matrix(e,f) == phi.matrix(e,f) - psi.matrix(e,f)  # check
            True
            sage: s == phi - psi  # indirect doctest
            True

        Other tests::

            sage: phi._sub_(Hom(M,N).zero()) == phi
            True
            sage: Hom(M,N).zero()._sub_(phi) == -phi
            True
            sage: phi._sub_(phi).is_zero()
            True
        """
        # No need for consistency checks since self and other are guaranteed
        # to have the same parents
        bases = self._common_bases(other)
        if bases is None:
            raise ValueError("no common pair of bases has been found to " +
                             "subtract {} from {}".format(other, self))
        # Subtraction at the matrix level:
        resu_mat = self._matrices[bases] - other._matrices[bases]
        if self._name is not None and other._name is not None:
            resu_name = self._name + '-' + other._name
        else:
            resu_name = None
        if self._latex_name is not None and other._latex_name is not None:
            resu_latex_name = self._latex_name + '-' + other._latex_name
        else:
            resu_latex_name = None
        return self.__class__(self.parent(), resu_mat, bases=bases,
                              name=resu_name, latex_name=resu_latex_name)

    def _lmul_(self, scalar):
        r"""
        Multiplication by ``scalar``.

        INPUT:

        - ``scalar`` -- element of the ring over which the parent of ``self``
          is a module

        OUTPUT: the homomorphism resulting from the multiplication of ``self``
        by ``scalar``

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]], name='phi',
            ....:             latex_name=r'\phi')
            sage: s = phi._lmul_(7) ; s
            Generic morphism:
              From: Rank-3 free module M over the Integer Ring
              To:   Rank-2 free module N over the Integer Ring
            sage: s.matrix(e,f)
            [-7 14  0]
            [35  7 14]
            sage: s == 7 * phi
            True
        """
        resu = self.__class__(self.parent(), 0)  # 0 = provisory value
        for bases, mat in self._matrices.items():
            resu._matrices[bases] = scalar * mat
        return resu

    #
    #  Other module methods
    #

    def __pos__(self):
        r"""
        Unary plus operator.

        OUTPUT: an exact copy of ``self``

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]], name='phi',
            ....:             latex_name=r'\phi')
            sage: s = phi.__pos__() ; s
            Generic morphism:
              From: Rank-3 free module M over the Integer Ring
              To:   Rank-2 free module N over the Integer Ring
            sage: s == +phi
            True
            sage: s == phi
            True
            sage: s is phi
            False
        """
        resu = self.__class__(self.parent(), 0, is_identity=self._is_identity)
                                           # 0 = provisory value
        for bases, mat in self._matrices.items():
            resu._matrices[bases] = +mat
        if self._name is not None:
            resu._name = '+' + self._name
        if self._latex_name is not None:
            resu._latex_name = '+' + self._latex_name
        return resu

    def __neg__(self):
        r"""
        Unary minus operator.

        OUTPUT: the homomorphism `-f`, where `f` is ``self``

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]], name='phi',
            ....:             latex_name=r'\phi')
            sage: s = phi.__neg__() ; s
            Generic morphism:
              From: Rank-3 free module M over the Integer Ring
              To:   Rank-2 free module N over the Integer Ring
            sage: s == -phi
            True
            sage: s.matrix()
            [ 1 -2  0]
            [-5 -1 -2]
            sage: s.matrix() == -phi.matrix()
            True
        """
        resu = self.__class__(self.parent(), 0)  # 0 = provisory value
        for bases, mat in self._matrices.items():
            resu._matrices[bases] = -mat
        if self._name is not None:
            resu._name = '-' + self._name
        if self._latex_name is not None:
            resu._latex_name = '-' + self._latex_name
        return resu

    #
    # Map methods
    #

    def _call_(
        self, element: FiniteRankFreeModuleElement
    ) -> FiniteRankFreeModuleElement:
        r"""
        Action of the homomorphism ``self`` on some free module element

        INPUT:

        - ``element`` -- element of the domain of ``self``

        OUTPUT: the image of ``element`` by ``self``

        EXAMPLES:

        Images of a homomorphism between two `\ZZ`-modules::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]], name='phi', latex_name=r'\phi')
            sage: v = M([1,2,3], basis=e, name='v')
            sage: w = phi(v) ; w
            Element phi(v) of the Rank-2 free module N over the Integer Ring
            sage: w.display()
            phi(v) = 3 f_0 + 13 f_1

        TESTS::

            sage: all(w[i] == sum(phi.matrix()[i,j]*v[j] for j in range(3)) for i in range(2))
            True
            sage: phi.matrix(e,f)
            [-1  2  0]
            [ 5  1  2]
            sage: phi(e[0]).display()
            phi(e_0) = -f_0 + 5 f_1
            sage: phi(e[1]).display()
            phi(e_1) = 2 f_0 + f_1
            sage: phi(e[2]).display()
            phi(e_2) = 2 f_1

        Image of an element that is not defined on the default basis::

            sage: a = M.automorphism()
            sage: a[0,2], a[1,0], a[2,1] = 1, -1, -1
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: v = M([1,2,3], basis=ep, name='v')
            sage: w = phi(v) ; w
            Element phi(v) of the Rank-2 free module N over the Integer Ring
            sage: w.display()
            phi(v) = -5 f_0 + 10 f_1
            sage: all(w[i] == sum(phi.matrix(ep,f)[i,j]*v[ep,j] for j in range(3)) for i in range(2))
            True

        Check of homomorphism properties::

            sage: phi(M.zero()) == N.zero()
            True
        """
        if self._is_identity:
            return element
        dom = self.parent().domain()
        sindex = dom._sindex
        codom = self.parent().codomain()
        basis_codom = codom.default_basis()
        # Search for a common basis to compute the result
        for basis in element._components:
            try:
                self.matrix(basis, basis_codom)
                basis_dom = basis
                break
            except ValueError:
                continue
        else:
            raise ValueError("no common basis found to evaluate the image " +
                             "of {} by {}".format(element,self))
        # Components of the result obtained by matrix multiplication
        mat = self.matrix(basis_dom, basis_codom)
        vcomp = element._components[basis_dom]
        tresu = []
        for i in range(codom.rank()):
            s = 0
            for j in range(dom.rank()):
                s += mat[i,j] * vcomp[[j+sindex]]
            tresu.append(s)
        # Name of the result
        if self._name is not None and element._name is not None:
            resu_name = self._name + '(' + element._name + ')'
        else:
            resu_name = None
        if self._latex_name is not None and element._latex_name is not None:
            resu_latex_name = self._latex_name + r'\left(' + \
                              element._latex_name + r'\right)'
        else:
            resu_latex_name = None
        # Creation of the result
        return codom(tresu, basis=basis_codom, name=resu_name,
                     latex_name=resu_latex_name)

    def is_injective(self):
        r"""
        Determine whether ``self`` is injective.

        OUTPUT:

        - ``True`` if ``self`` is an injective homomorphism and ``False``
          otherwise

        EXAMPLES:

        Homomorphisms between two `\ZZ`-modules::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]])
            sage: phi.matrix(e,f)
            [-1  2  0]
            [ 5  1  2]
            sage: phi.is_injective()
            False

        Indeed, phi has a non trivial kernel::

            sage: phi(4*e[0] + 2*e[1] - 11*e[2]).display()
            0

        An injective homomorphism::

            sage: psi = N.hom(M, [[1,-1], [0,3], [4,-5]])
            sage: psi.matrix(f,e)
            [ 1 -1]
            [ 0  3]
            [ 4 -5]
            sage: psi.is_injective()
            True

        Of course, the identity endomorphism is injective::

            sage: End(M).one().is_injective()
            True
            sage: End(N).one().is_injective()
            True
        """
        # Some matrix representation is picked at random:
        matrix_rep = next(iter(self._matrices.values()))
        return not matrix_rep.right_kernel().rank()

    def is_surjective(self):
        r"""
        Determine whether ``self`` is surjective.

        OUTPUT:

        - ``True`` if ``self`` is a surjective homomorphism and ``False``
          otherwise

        EXAMPLES:

        This method has not been implemented yet::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]])
            sage: phi.is_surjective()
            Traceback (most recent call last):
            ...
            NotImplementedError: FiniteRankFreeModuleMorphism.is_surjective()
             has not been implemented yet

        except for the identity endomorphism (!)::

            sage: End(M).one().is_surjective()
            True
            sage: End(N).one().is_surjective()
            True
        """
        if self._is_identity:
            return True
        raise NotImplementedError(
                              "FiniteRankFreeModuleMorphism.is_surjective() " +
                              "has not been implemented yet")
    #
    # Morphism methods
    #

    def is_identity(self):
        r"""
        Check whether ``self`` is the identity morphism.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 2, name='M')
            sage: e = M.basis('e')
            sage: phi = M.endomorphism([[1,0], [0,1]])
            sage: phi.is_identity()
            True
            sage: (phi+phi).is_identity()
            False
            sage: End(M).zero().is_identity()
            False
            sage: a = M.automorphism() ; a[0,1], a[1,0] = 1, -1
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: phi = M.endomorphism([[1,0], [0,1]], basis=ep)
            sage: phi.is_identity()
            True

        Example illustrating that the identity can be constructed from a
        matrix that is not the identity one, provided that it is relative to
        different bases::

            sage: phi = M.hom(M, [[0,1], [-1,0]], bases=(ep,e))
            sage: phi.is_identity()
            True

        Of course, if we ask for the matrix in a single basis, it is the
        identity matrix::

            sage: phi.matrix(e)
            [1 0]
            [0 1]
            sage: phi.matrix(ep)
            [1 0]
            [0 1]
        """
        if self._is_identity:
            return True
        # The identity must be an endomorphism:
        fmodule = self.domain()
        if fmodule != self.codomain():
            return False
        # Some basis in which ``self`` has a representation is picked at
        # random and the test is performed on the images of the basis
        # elements:
        basis = list(self._matrices)[0][0]
        for i in fmodule.irange():
            if self(basis[i]) != basis[i]:
                return False
        self._is_identity = True
        return True

    #
    # End of Morphism methods
    #

    def _modules_and_bases(self, basis1=None, basis2=None):
        r"""
        Return domain, codomain, domain basis, and codomain basis.

        This method implements default argument handling for methods
        :meth:`matrix` and :meth:`display`.

        INPUT:

        - ``basis1`` -- (default: ``None``) basis of the domain of ``self``; if
          none is provided, the domain's default basis is assumed
        - ``basis2`` -- (default: ``None``) basis of the codomain of ``self``;
          if none is provided, ``basis2`` is set to ``basis1`` if ``self`` is
          an endomorphism, otherwise, ``basis2`` is set to the codomain's
          default basis.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e'); f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]])
            sage: phi._modules_and_bases()
            (Rank-3 free module M over the Integer Ring,
             Rank-2 free module N over the Integer Ring,
             Basis (e_0,e_1,e_2) on the Rank-3 free module M over the Integer Ring,
             Basis (f_0,f_1) on the Rank-2 free module N over the Integer Ring)
            sage: a = M.automorphism(matrix=[[-1,0,0],[0,1,2],[0,1,3]], basis=e)
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: phi._modules_and_bases(ep)
            (Rank-3 free module M over the Integer Ring,
             Rank-2 free module N over the Integer Ring,
             Basis (ep_0,ep_1,ep_2) on the Rank-3 free module M over the Integer Ring,
             Basis (f_0,f_1) on the Rank-2 free module N over the Integer Ring)
        """
        fmodule1 = self.domain()
        fmodule2 = self.codomain()
        if basis1 is None:
            basis1 = fmodule1.default_basis()
        elif basis1 not in fmodule1.bases():
            raise TypeError(str(basis1) + " is not a basis on the " +
                            str(fmodule1) + ".")
        if basis2 is None:
            if self.is_endomorphism():
                basis2 = basis1
            else:
                basis2 = fmodule2.default_basis()
        elif basis2 not in fmodule2.bases():
            raise TypeError(str(basis2) + " is not a basis on the " +
                            str(fmodule2) + ".")
        return fmodule1, fmodule2, basis1, basis2

    def matrix(self, basis1=None, basis2=None):
        r"""
        Return the matrix of ``self`` w.r.t to a pair of bases.

        If the matrix is not known already, it is computed from the matrix in
        another pair of bases by means of the change-of-basis formula.

        INPUT:

        - ``basis1`` -- (default: ``None``) basis of the domain of ``self``; if
          none is provided, the domain's default basis is assumed
        - ``basis2`` -- (default: ``None``) basis of the codomain of ``self``;
          if none is provided, ``basis2`` is set to ``basis1`` if ``self`` is
          an endomorphism, otherwise, ``basis2`` is set to the codomain's
          default basis.

        OUTPUT:

        - the matrix representing the homomorphism ``self`` w.r.t
          to bases ``basis1`` and ``basis2``; more precisely, the columns of
          this matrix are formed by the components w.r.t. ``basis2`` of
          the images of the elements of ``basis1``.

        EXAMPLES:

        Matrix of a homomorphism between two `\ZZ`-modules::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]])
            sage: phi.matrix()     # default bases
            [-1  2  0]
            [ 5  1  2]
            sage: phi.matrix(e, f)  # given bases
            [-1  2  0]
            [ 5  1  2]
            sage: type(phi.matrix())                                                    # needs sage.libs.flint
            <class 'sage.matrix.matrix_integer_dense.Matrix_integer_dense'>

        Matrix in bases different from those in which the homomorphism has
        been defined::

            sage: a = M.automorphism(matrix=[[-1,0,0],[0,1,2],[0,1,3]], basis=e)
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: b = N.automorphism(matrix=[[3,5],[4,7]], basis=f)
            sage: fp = f.new_basis(b, 'fp', latex_symbol="f'")
            sage: phi.matrix(ep, fp)
            [ 32  -1 -12]
            [-19   1   8]

        Check of the change-of-basis formula::

            sage: phi.matrix(ep, fp) == (b^(-1)).matrix(f) * phi.matrix(e,f) * a.matrix(e)
            True

        Single change of basis::

            sage: phi.matrix(ep, f)
            [ 1  2  4]
            [-5  3  8]
            sage: phi.matrix(ep,f) == phi.matrix(e,f) * a.matrix(e)
            True
            sage: phi.matrix(e, fp)
            [-32   9 -10]
            [ 19  -5   6]
            sage: phi.matrix(e, fp) == (b^(-1)).matrix(f) * phi.matrix(e,f)
            True

        Matrix of an endomorphism::

            sage: phi = M.endomorphism([[1,2,3], [4,5,6], [7,8,9]], basis=ep)
            sage: phi.matrix(ep)
            [1 2 3]
            [4 5 6]
            [7 8 9]
            sage: phi.matrix(ep,ep)  # same as above
            [1 2 3]
            [4 5 6]
            [7 8 9]
            sage: phi.matrix()  # matrix w.r.t to the module's default basis
            [  1  -3   1]
            [-18  39 -18]
            [-25  54 -25]
        """
        from sage.matrix.constructor import matrix
        fmodule1, fmodule2, basis1, basis2 = self._modules_and_bases(basis1, basis2)
        if (basis1, basis2) not in self._matrices:
            if self._is_identity:
                # The identity endomorphism
                # -------------------------
                if basis1 == basis2:
                    # the matrix is the identity matrix:
                    ring = fmodule1.base_ring()
                    zero = ring.zero()
                    one = ring.one()
                    size = fmodule1.rank()
                    mat = []
                    for i in range(size):
                        row = [zero]*size
                        row[i] = one
                        mat.append(row)
                else:
                    # the matrix is the change-of-basis matrix:
                    change = fmodule1.change_of_basis(basis1, basis2)
                    mat = [[change[[i,j]] for j in fmodule1.irange()]
                                                    for i in fmodule1.irange()]
                self._matrices[(basis1, basis2)] = matrix(mat)
            else:
                # Generic homomorphism
                # --------------------
                b1_list = [bases[0] for bases in self._matrices]
                b2_list = [bases[1] for bases in self._matrices]
                if basis1 in b1_list:
                    for b2 in b2_list:
                        if (basis2, b2) in fmodule2._basis_changes:
                            nb2 = b2
                            break
                    else:
                        raise ValueError("no start basis could be found for " +
                                        "applying the change-of-basis formula")
                    change2 = fmodule2._basis_changes[(basis2, nb2)]
                    mat2 = matrix( [[change2[[i,j]] for j in fmodule2.irange()]
                                                  for i in fmodule2.irange()] )
                    self._matrices[(basis1, basis2)] = \
                                            mat2 * self._matrices[(basis1,nb2)]
                elif basis2 in b2_list:
                    for b1 in b1_list:
                        if (b1, basis1) in fmodule1._basis_changes:
                            nb1 = b1
                            break
                    else:
                        raise ValueError("no start basis could be found for " +
                                        "applying the change-of-basis formula")
                    change1 = fmodule1._basis_changes[(nb1, basis1)]
                    mat1 = matrix( [[change1[[i,j]] for j in fmodule1.irange()]
                                                  for i in fmodule1.irange()] )
                    self._matrices[(basis1, basis2)] = \
                                            self._matrices[(nb1,basis2)] * mat1
                else: # most general change-of-basis formula
                    for (b1, b2) in self._matrices:
                        if (b1, basis1) in fmodule1._basis_changes and \
                           (basis2, b2) in fmodule2._basis_changes:
                            nb1, nb2 = b1, b2
                            break
                    else:
                        raise ValueError("no start basis could be found for " +
                                        "applying the change-of-basis formula")
                    change1 = fmodule1._basis_changes[(nb1, basis1)]
                    change2 = fmodule2._basis_changes[(basis2, nb2)]
                    mat1 = matrix( [[change1[[i,j]] for j in fmodule1.irange()]
                                                  for i in fmodule1.irange()] )
                    mat2 = matrix( [[change2[[i,j]] for j in fmodule2.irange()]
                                                  for i in fmodule2.irange()] )
                    self._matrices[(basis1, basis2)] = \
                                        mat2 * self._matrices[(nb1,nb2)] * mat1
        return self._matrices[(basis1, basis2)]

    def _common_bases(self, other):
        r"""
        Return a pair of bases in which ``self`` and ``other`` have a known
        matrix representation.

        INPUT:

        - ``other`` -- another homomorphism in the same hom-set

        OUTPUT:

        - a pair of bases in which ``self`` and ``other`` have a known
          matrix representation.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e') ; f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]])
            sage: psi = M.hom(N, [[1,1,0], [4,1,3]])
            sage: phi._common_bases(psi) # matrices of phi and psi both defined on (e,f)
            (Basis (e_0,e_1,e_2) on the Rank-3 free module M over the Integer Ring,
             Basis (f_0,f_1) on the Rank-2 free module N over the Integer Ring)
            sage: a = M.automorphism() ; a[0,2], a[1,0], a[2,1] = 1, -1, -1
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: psi = M.hom(N, [[1,1,0], [4,1,3]], bases=(ep,f))
            sage: phi._common_bases(psi) # matrix of psi w.r.t. (e,f) computed
            (Basis (e_0,e_1,e_2) on the Rank-3 free module M over the Integer Ring,
             Basis (f_0,f_1) on the Rank-2 free module N over the Integer Ring)
            sage: psi = M.hom(N, [[1,1,0], [4,1,3]], bases=(ep,f))
            sage: psi._common_bases(phi) # matrix of phi w.r.t. (ep,f) computed
            (Basis (ep_0,ep_1,ep_2) on the Rank-3 free module M over the Integer Ring,
             Basis (f_0,f_1) on the Rank-2 free module N over the Integer Ring)
        """
        resu = None
        for bases in self._matrices:
            try:
                other.matrix(*bases)
                resu = bases
                break
            except ValueError:
                continue
        if resu is None:
            for bases in other._matrices:
                try:
                    self.matrix(*bases)
                    resu = bases
                    break
                except ValueError:
                    continue
        return resu

    def display(self, basis1=None, basis2=None):
        r"""
        Display ``self`` as a matrix w.r.t to a pair of bases.

        If the matrix is not known already, it is computed from the matrix in
        another pair of bases by means of the change-of-basis formula.

        INPUT:

        - ``basis1`` -- (default: ``None``) basis of the domain of ``self``; if
          none is provided, the domain's default basis is assumed
        - ``basis2`` -- (default: ``None``) basis of the codomain of ``self``;
          if none is provided, ``basis2`` is set to ``basis1`` if ``self`` is
          an endomorphism, otherwise, ``basis2`` is set to the codomain's
          default basis.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: N = FiniteRankFreeModule(ZZ, 2, name='N')
            sage: e = M.basis('e'); f = N.basis('f')
            sage: phi = M.hom(N, [[-1,2,0], [5,1,2]])
            sage: phi.display()     # default bases
                e_0 e_1 e_2
            f_0⎛ -1   2   0⎞
            f_1⎝  5   1   2⎠
            sage: phi.display(e, f)  # given bases
                e_0 e_1 e_2
            f_0⎛ -1   2   0⎞
            f_1⎝  5   1   2⎠

        Matrix of an endomorphism::

            sage: a = M.automorphism(matrix=[[-1,0,0],[0,1,2],[0,1,3]], basis=e)
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: phi = M.endomorphism([[1,2,3], [4,5,6], [7,8,9]], basis=ep)
            sage: phi.display(ep)
                 ep_0 ep_1 ep_2
            ep_0⎛   1    2    3⎞
            ep_1⎜   4    5    6⎟
            ep_2⎝   7    8    9⎠
            sage: phi.display(ep, ep)  # same as above
                 ep_0 ep_1 ep_2
            ep_0⎛   1    2    3⎞
            ep_1⎜   4    5    6⎟
            ep_2⎝   7    8    9⎠
            sage: phi.display()  # matrix w.r.t to the module's default basis
                e_0 e_1 e_2
            e_0⎛  1  -3   1⎞
            e_1⎜-18  39 -18⎟
            e_2⎝-25  54 -25⎠
        """
        from sage.misc.latex import latex
        from .format_utilities import is_atomic, FormattedExpansion
        fmodule1, fmodule2, basis1, basis2 = self._modules_and_bases(basis1, basis2)
        matrix = self.matrix(basis1, basis2)
        if all(element._name for element in basis1):
            basis1_names = [element._name for element in basis1]
        else:
            basis1_names = None
        if all(element._name for element in basis2):
            basis2_names = [element._name for element in basis2]
        else:
            basis2_names = None
        resu_txt = matrix.str(unicode=True,
                              top_border=basis1_names,
                              left_border=basis2_names)
        resu_latex = latex(matrix)
        return FormattedExpansion(resu_txt, resu_latex)


class FiniteRankFreeModuleEndomorphism(FiniteRankFreeModuleMorphism):
    r"""
    Endomorphism of a free module of finite rank over a commutative ring.

    An instance of this class is an endomorphism

    .. MATH::

        \phi:\ M \longrightarrow M,

    where `M` is a free module of finite rank over a commutative ring `R`.

    This is a Sage *element* class, the corresponding *parent* class being
    :class:`~sage.tensor.modules.free_module_homset.FreeModuleEndset`.

    INPUT:

    - ``parent`` -- Hom-set Hom(M,M) to which the endomorphism belongs
    - ``matrix_rep`` -- matrix representation of the endomorphism with
      respect to the basis ``bases``; this entry can actually
      be any material from which a matrix of size rank(N)*rank(M) of
      elements of `R` can be constructed; the *columns* of the matrix give
      the images of the basis of `M` (see the convention in the example below)
    - ``bases`` -- (default: ``None``) pair ``(basis_domain, basis_codomain)``
      defining the matrix representation, ``basis_domain`` and ``basis_codomain``
      being two bases (typically the same) of the same module `M`; if ``None``,
      the default basis of `M` is used for both.
    - ``name`` -- (default: ``None``) string; name given to the endomorphism
    - ``latex_name`` -- (default: ``None``) string; LaTeX symbol to denote the
      endomorphism. If ``None``, ``name`` will be used.
    - ``is_identity`` -- boolean (default: ``False``); determines whether the
      constructed object is the identity endomorphism. If set to ``True``,
      then the entry ``matrix_rep`` is not used.

    EXAMPLES:

    The identity endomorphism::

        sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
        sage: e = M.basis('e')
        sage: Id = End(M).one(); Id
        Identity endomorphism of Rank-3 free module M over the Integer Ring
        sage: Id.parent()
        Set of Morphisms from Rank-3 free module M over the Integer Ring
         to Rank-3 free module M over the Integer Ring
         in Category of finite dimensional modules over Integer Ring
        sage: Id.parent() is End(M)
        True

    The matrix of ``Id`` with respect to the basis ``e`` is of course the identity
    matrix::

        sage: Id.matrix(e)
        [1 0 0]
        [0 1 0]
        [0 0 1]

    The identity acting on a module element::

        sage: v = M([-2,1,4], basis=e, name='v'); v.display()
        v = -2 e_0 + e_1 + 4 e_2
        sage: Id(v) is v
        True
    """
    def __init__(self, parent, matrix_rep, bases=None, name=None,
                 latex_name=None, is_identity=False):
        r"""
        TESTS::

            sage: from sage.tensor.modules.free_module_morphism import FiniteRankFreeModuleEndomorphism
            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')

        Generic endomorphism::

            sage: phi = FiniteRankFreeModuleEndomorphism(End(M),
            ....:                                        [[1,0,-3], [2,1,4], [7,8,9]],
            ....:                                        name='phi', latex_name=r'\phi')
            sage: phi
            Generic endomorphism of Rank-3 free module M over the Integer Ring

        Identity endomorphism::

            sage: phi = FiniteRankFreeModuleEndomorphism(End(M), 'whatever',
            ....:                                        is_identity=True)
            sage: phi
            Identity endomorphism of Rank-3 free module M over the Integer Ring
            sage: phi.matrix(e)
            [1 0 0]
            [0 1 0]
            [0 0 1]
            sage: latex(phi)
            \mathrm{Id}
        """
        from sage.misc.constant_function import ConstantFunction

        fmodule = parent.domain()
        if bases is None:
            def_basis = fmodule.default_basis()
            if def_basis is None:
                raise ValueError("the {} has no default ".format(fmodule) +
                                 "basis")
            bases = (def_basis, def_basis)
        else:
            bases = tuple(bases)  # insures bases is a tuple
            if len(bases) != 2:
                raise TypeError("the argument bases must contain 2 bases")
            if bases[0] not in fmodule.bases():
                raise TypeError("{} is not a basis on the {}".format(bases[0],
                                                                     fmodule))
            if bases[1] not in fmodule.bases():
                raise TypeError("{} is not a basis on the {}".format(bases[1],
                                                                     fmodule))
        if not is_identity:
            # Construction of a generic endomorphism
            if isinstance(matrix_rep, ConstantFunction):
                # the zero morphism
                if matrix_rep().is_zero():
                    matrix_rep = 0
            if bases[0] == bases[1] and matrix_rep == 1:
                is_identity = True
        if is_identity:
            # Construction of the identity endomorphism
            if bases[0] != bases[1]:
                raise TypeError("the two bases must coincide for " +
                                "constructing the identity endomorphism.")
            matrix_rep = 1
            if name is None:
                name = 'Id'
            if latex_name is None and name == 'Id':
                latex_name = r'\mathrm{Id}'
        FiniteRankFreeModuleMorphism.__init__(self, parent, matrix_rep, bases,
                                              name, latex_name)
        if is_identity:
            self._is_identity = True
            self._repr_type_str = 'Identity'

    def _some_matrix(self):
        r"""
        Return the matrix of the endomorphism ``self`` w.r.t. some basis.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: a = M.automorphism(matrix=[[-1,0,0],[0,1,2],[0,1,3]], basis=e)
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: phi = M.endomorphism([[1,2,3], [4,5,6], [7,8,9]], basis=ep)
            sage: phi._some_matrix()
            [1 2 3]
            [4 5 6]
            [7 8 9]
        """
        for (basis1, basis2), matrix in self._matrices.items():
            if basis1 == basis2:
                return matrix
        for basis in self.domain().bases():
            try:
                return self.matrix(basis)
            except ValueError:
                pass
        return self.matrix()

    @lazy_attribute
    def characteristic_polynomial(self):
        r"""
        Return the characteristic polynomial of ``self``.

        :meth:`characteristic_polynomial` and :meth:`charpoly` are the same method.

        INPUT:

        - ``var`` -- string (default: ``'x'``); a variable name

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: a = M.automorphism(matrix=[[-1,0,0],[0,1,2],[0,1,3]], basis=e)
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: phi = M.endomorphism([[1,2,3], [4,5,6], [7,8,9]], basis=ep)
            sage: phi.matrix(e)
            [  1  -3   1]
            [-18  39 -18]
            [-25  54 -25]
            sage: phi.characteristic_polynomial()
            x^3 - 15*x^2 - 18*x
            sage: phi.charpoly()
            x^3 - 15*x^2 - 18*x
            sage: phi.charpoly('T')
            T^3 - 15*T^2 - 18*T
        """
        return self._some_matrix().characteristic_polynomial

    charpoly = characteristic_polynomial

    @lazy_attribute
    def det(self):
        r"""
        Return the determinant of ``self``.

        OUTPUT:

        - element of the base ring of the modules on which ``self`` is defined,
          equal to the determinant of ``self``.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: a = M.automorphism(matrix=[[-1,0,0],[0,1,2],[0,1,3]], basis=e)
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: phi = M.endomorphism([[1,2,3], [4,5,6], [7,8,9]], basis=ep)
            sage: phi.matrix(e)
            [  1  -3   1]
            [-18  39 -18]
            [-25  54 -25]
            sage: phi.det()
            0
            sage: det(phi)
            0
        """
        return self._some_matrix().det

    determinant = det

    @lazy_attribute
    def fcp(self):
        r"""
        Return the factorization of the characteristic polynomial of ``self``.

        INPUT:

        - ``var`` -- string (default: ``'x'``); a variable name

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: a = M.automorphism(matrix=[[-1,0,0],[0,1,2],[0,1,3]], basis=e)
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: phi = M.endomorphism([[1,2,3], [4,5,6], [7,8,9]], basis=ep)
            sage: phi.matrix(e)
            [  1  -3   1]
            [-18  39 -18]
            [-25  54 -25]
            sage: phi.fcp()                                                             # needs sage.libs.pari
            x * (x^2 - 15*x - 18)
            sage: phi.fcp('T')                                                          # needs sage.libs.pari
            T * (T^2 - 15*T - 18)
        """
        return self._some_matrix().fcp

    @lazy_attribute
    def minimal_polynomial(self):
        r"""
        Return the minimal polynomial of ``self``.

        :meth:`minimal_polynomial` and :meth:`minpoly` are the same method.

        INPUT:

        - ``var`` -- string (default: ``'x'``); a variable name

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: a = M.automorphism(matrix=[[-1,0,0],[0,1,2],[0,1,3]], basis=e)
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: phi = M.endomorphism([[1,2,3], [4,5,6], [7,8,9]], basis=ep)
            sage: phi.matrix(e)
            [  1  -3   1]
            [-18  39 -18]
            [-25  54 -25]
            sage: phi.minpoly()                                                         # needs sage.libs.pari
            x^3 - 15*x^2 - 18*x
            sage: phi.minimal_polynomial()                                              # needs sage.libs.pari
            x^3 - 15*x^2 - 18*x
            sage: phi.minimal_polynomial('T')                                           # needs sage.libs.pari
            T^3 - 15*T^2 - 18*T
        """
        return self._some_matrix().minimal_polynomial

    minpoly = minimal_polynomial

    @lazy_attribute
    def trace(self):
        r"""
        Return the trace of ``self``.

        OUTPUT:

        - element of the base ring of the modules on which ``self`` is defined,
          equal to the trace of ``self``.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: a = M.automorphism(matrix=[[-1,0,0],[0,1,2],[0,1,3]], basis=e)
            sage: ep = e.new_basis(a, 'ep', latex_symbol="e'")
            sage: phi = M.endomorphism([[1,2,3], [4,5,6], [7,8,9]], basis=ep)
            sage: phi.matrix(e)
            [  1  -3   1]
            [-18  39 -18]
            [-25  54 -25]
            sage: phi.trace()
            15
            sage: id = M.identity_map()
            sage: id.trace()
            3
        """
        return self._some_matrix().trace



# sage_setup: distribution = sagemath-flint
# sage.doctest: needs sage.graphs sage.modules
r"""
Auslander-Reiten Quivers
"""

# ****************************************************************************
#  Copyright (C) 2024 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.sets_cat import Sets
from sage.combinat.free_module import CombinatorialFreeModule
from sage.combinat.root_system.cartan_type import CartanType
from sage.graphs.digraph import DiGraph
from sage.misc.cachefunc import cached_method
from sage.rings.integer_ring import ZZ
from sage.sets.family import Family
from sage.structure.element import Element
from sage.structure.global_options import GlobalOptions
from sage.structure.parent import Parent
from sage.structure.richcmp import richcmp
from sage.structure.unique_representation import UniqueRepresentation


class AuslanderReitenQuiver(UniqueRepresentation, Parent):
    r"""
    The Auslander-Reiten quiver.

    Let `Q = (Q_0, Q_1)` be a finite acyclic quiver. The
    *Auslander-Reiten quiver* (AR quiver) `\Gamma_Q` is the quiver
    whose vertices correspond to the indecompositible modules of `Q`
    (equivalently its path algebra over an algebraically closed field)
    and edges are irreducible morphisms.

    In this implementation, we denote the vertices of `\Gamma_Q` as
    certain pairs `\langle v, k \rangle`, where `v \in Q_0` and
    `k \in \ZZ \setminus \{0\}` is called the *level*. When `k > 0`
    (resp. `k < 0`), then it corresponds to a preprojective (resp.
    postinjective) module. When the quiver is a finite type Dynkin
    quiver, we consider all modules to be preprojectives and denoted
    by a positive level.

    .. NOTE::

        We use the terminology *postinjective* instead of *preinjective*
        given that they follow from injectives by AR translation.

    ALGORITHM:

    We compute the dimension vectors of a projective `\langle v, 1 \rangle`
    by counting the number of (directed) paths `u \to v` in `Q`. We then
    proceed inductively to compute all of the dimension vectors of level
    `k` by using the translation equation

    .. MATH::

        dim \langle v, k-1 \rangle + \dim \langle v, k \rangle
        = \sum_{u,k'} \dim \langle u, k' \rangle,

    where the sum is over all paths from `\langle v, k-1 \rangle` to
    `\langle v, k \rangle` in `\Gamma_Q`. More specifically, for each edge
    `(u, v, \ell) \in Q_1` (resp. `(v, u, \ell) \in Q_1`), we have
    `\langle u, k-1 \rangle` (resp. `\langle u, k \rangle`) in the sum
    (assuming the node is in the AR quiver).

    The algorithm for postinjectives is dual to the above.

    .. TODO::

        This only is implemented for the preprojectives and postinjectives
        when the quiver is not a finite type Dynkin quiver.

    .. TODO::

        Implement this for general Artinian algebras.

    EXAMPLES:

    We create the AR quivers for finite type `A_3` Dynkin quivers::

        sage: DA = DiGraph([[1, 2], [2, 3]])
        sage: AR = DA.auslander_reiten_quiver()
        sage: AR.digraph().edges(labels=False)
        [(<1, 1>, <2, 2>), (<2, 1>, <1, 1>), (<2, 1>, <3, 2>), (<3, 1>, <2, 1>),
         (<2, 2>, <3, 3>), (<3, 2>, <2, 2>)]

        sage: DA = DiGraph([[1, 2], [3, 2]])
        sage: AR = DA.auslander_reiten_quiver()
        sage: AR.digraph().edges(labels=False)
        [(<1, 1>, <2, 2>), (<2, 1>, <1, 1>), (<2, 1>, <3, 1>), (<3, 1>, <2, 2>),
         (<2, 2>, <1, 2>), (<2, 2>, <3, 2>)]

        sage: DA = DiGraph([[2, 1], [2, 3]])
        sage: AR = DA.auslander_reiten_quiver()
        sage: AR.digraph().edges(labels=False)
        [(<1, 1>, <2, 1>), (<2, 1>, <1, 2>), (<2, 1>, <3, 2>), (<3, 1>, <2, 1>),
         (<1, 2>, <2, 2>), (<3, 2>, <2, 2>)]

        sage: DA = DiGraph([[2, 1], [3, 2]])
        sage: AR = DA.auslander_reiten_quiver()
        sage: AR.digraph().edges(labels=False)
        [(<1, 1>, <2, 1>), (<2, 1>, <3, 1>), (<2, 1>, <1, 2>), (<3, 1>, <2, 2>),
         (<1, 2>, <2, 2>), (<2, 2>, <1, 3>)]

    An example for the type `D_5` Dynkin quiver::

        sage: DD = DiGraph([[5,3], [4,3], [3,2], [2,1]])
        sage: AR = DD.auslander_reiten_quiver()
        sage: AR
        Auslander-Reiten quiver of a ['D', 5] Dynkin quiver
        sage: len(list(DD))
        5

    An `E_8` Dynkin quiver::

        sage: DE = DiGraph([[8,7], [7,6], [5,6], [5,3], [3,4], [3,2], [2,1]])
        sage: AR = DE.auslander_reiten_quiver()
        sage: AR
        Auslander-Reiten quiver of a ['E', 8] Dynkin quiver
        sage: len(list(AR))
        120
        sage: len(list(RootSystem(['E', 8]).root_lattice().positive_roots()))
        120

    The Kronecker quiver::

        sage: D = DiGraph([[1,2,'a'], [1,2,'b']], multiedges=True)
        sage: AR = D.auslander_reiten_quiver()
        sage: for i in range(1, 5):
        ....:     for v in D.vertices():
        ....:         pp = AR(v, i)
        ....:         pi = AR(v, -i)
        ....:         print(pp, pp.dimension_vector(), "  ", pi, pi.dimension_vector())
        <1, 1> v1 + 2*v2      <1, -1> v1
        <2, 1> v2             <2, -1> 2*v1 + v2
        <1, 2> 3*v1 + 4*v2    <1, -2> 3*v1 + 2*v2
        <2, 2> 2*v1 + 3*v2    <2, -2> 4*v1 + 3*v2
        <1, 3> 5*v1 + 6*v2    <1, -3> 5*v1 + 4*v2
        <2, 3> 4*v1 + 5*v2    <2, -3> 6*v1 + 5*v2
        <1, 4> 7*v1 + 8*v2    <1, -4> 7*v1 + 6*v2
        <2, 4> 6*v1 + 7*v2    <2, -4> 8*v1 + 7*v2
    """
    @staticmethod
    def __classcall_private__(cls, quiver):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: D = DiGraph([[1,2], [2,3], [3,1]])
            sage: D.auslander_reiten_quiver()
            Traceback (most recent call last):
            ...
            ValueError: the quiver must not have cycles
        """
        if quiver.has_loops() or not quiver.is_directed_acyclic():
            raise ValueError("the quiver must not have cycles")
        quiver = quiver.copy(immutable=True)
        return super().__classcall__(cls, quiver)

    def __init__(self, quiver):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: DE = DiGraph([[7,6], [6,5], [5,3], [3,4], [2,3], [1,2]])
            sage: AR = DE.auslander_reiten_quiver()
            sage: TestSuite(AR).run()

            sage: D = DiGraph([[1,2], [3,4]])
            sage: AR = D.auslander_reiten_quiver()
            sage: TestSuite(AR).run()

            sage: D = DiGraph([[1,2,'a'], [1,2,'b']], multiedges=True)
            sage: AR = D.auslander_reiten_quiver()
            sage: TestSuite(AR).run()
        """
        self._quiver = quiver
        self._top_sort = quiver.topological_sort()
        self._dim_vec_space = CombinatorialFreeModule(ZZ, quiver.vertices(), prefix='v', bracket=False)
        self._max_level = float('inf')

        dynkin_type = detect_dynkin_quiver(quiver)
        if dynkin_type is not None:
            self._cartan_type = dynkin_type
        self._is_finite = dynkin_type is not None
        cat = Sets().Enumerated().Finite() if self._is_finite else Sets().Infinite()
        super().__init__(category=cat)

    def _repr_(self) -> str:
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: DE = DiGraph([[5,6], [5,3], [3,4], [2,3], [1,2]])
            sage: DE.auslander_reiten_quiver()
            Auslander-Reiten quiver of a ['E', 6] Dynkin quiver

            sage: D = DiGraph([[1,2,'a'], [1,2,'b']], multiedges=True)
            sage: D.auslander_reiten_quiver()
            Auslander-Reiten quiver of Multi-digraph on 2 vertices
        """
        if self._is_finite:
            return "Auslander-Reiten quiver of a {} Dynkin quiver".format(self._cartan_type)
        return "Auslander-Reiten quiver of {}".format(self._quiver)

    # add options to class
    class options(GlobalOptions):
        r"""
        Set and display the global options for Auslander-Reiten quivers.
        If no parameters are set, then the function returns a copy of the
        options dictionary.

        The ``options`` to partitions can be accessed as the method
        :obj:`AuslanderReitenQuiver.options` of
        :class:`~sage.quivers.ar_quiver.AuslanderReitenQuiver`.

        @OPTIONS@

        EXAMPLES::

            sage: D = DiGraph([[1,2,'a'], [1,2,'b']], multiedges=True)
            sage: AR = D.auslander_reiten_quiver()
            sage: node = AR(2, 2)
            sage: latex(node)
            \left\langle 2, 2 \right\rangle
            sage: AR.options.latex = "dimension_vector"
            sage: latex(node)
            2 v_{1} + 3 v_{2}
            sage: AR.options.latex = "both"
            sage: latex(node)
            \begin{gathered} \left\langle 2, 2 \right\rangle \\ 2 v_{1} + 3 v_{2} \end{gathered}
            sage: AR.options._reset()
        """
        NAME = 'AuslanderReitenQuiver'
        module = 'sage.quivers.ar_quiver'
        latex = dict(default='node',
                     description='Specifies how nodes of the AR quiver should be latexed',
                     values=dict(node='latex as the node description',
                                 dimension_vector='latex as the dimension vector',
                                 both='latex as both'),
                     case_sensitive=False)

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: DE = DiGraph([[7,8], [7,6], [5,6], [3,5], [4,3], [2,3], [1,2]])
            sage: AR = DE.auslander_reiten_quiver()
            sage: AR._an_element_()
            <1, 1>

            sage: D = DiGraph([[1,2,'a'], [1,2,'b']], multiedges=True)
            sage: AR = D.auslander_reiten_quiver()
            sage: AR._an_element_()
            <1, 1>
        """
        return next(iter(self.projectives()))

    def quiver(self):
        r"""
        Return the quiver defining ``self``.

        EXAMPLES::

            sage: DE = DiGraph([[7,8], [7,6], [5,6], [3,5], [4,3], [2,3], [1,2]])
            sage: AR = DE.auslander_reiten_quiver()
            sage: AR.quiver() == DE
            True
        """
        return self._quiver

    def projectives(self):
        r"""
        Return the projectives of ``self``.

        EXAMPLES::

            sage: D = DiGraph([[1,2,'a'], [1,2,'b']], multiedges=True)
            sage: AR = D.auslander_reiten_quiver()
            sage: AR.projectives()
            Finite family {1: <1, 1>, 2: <2, 1>}
        """
        return Family({v: self.element_class(self, v, 1) for v in self._quiver.vertex_iterator()})

    @cached_method
    def simples(self):
        r"""
        Return the simples of ``self``.

        EXAMPLES::

            sage: DE = DiGraph([[7,8], [7,6], [5,6], [3,5], [4,3], [2,3], [1,2]])
            sage: AR = DE.auslander_reiten_quiver()
            sage: AR.simples()
            Finite family {1: <1, 15>,  2: <1, 14>,  3: <8, 4>,  4: <4, 15>,
                           5: <8, 3>,  6: <6, 1>,  7: <7, 15>,  8: <8, 1>}
        """
        ret = {}
        for elt in self:
            supp = elt.dimension_vector().support()
            if len(supp) != 1:
                continue
            ret[next(iter(supp))] = elt
        return Family(ret)

    def injectives(self):
        r"""
        Return the injectives of ``self``.

        EXAMPLES::

            sage: DE = DiGraph([[7,6], [6,5], [5,3], [4,3], [2,3], [1,2]])
            sage: AR = DE.auslander_reiten_quiver()
            sage: AR.injectives()
            Finite family {1: <1, 9>, 2: <2, 9>, 3: <3, 9>, 4: <4, 9>,
                           5: <5, 9>, 6: <6, 9>, 7: <7, 9>}
        """
        if self._is_finite:
            self.digraph()  # sets self._injective attribute
            return self._injectives
        return Family({v: self(v, -1) for v in self._quiver.vertex_iterator()})

    def _digraph_set_latex_options(self, G):
        """
        Set the latex options of the digraph ``G``.

        EXAMPLES::

            sage: D = DiGraph([[1,2,'a'], [1,2,'b']], multiedges=True)
            sage: AR = D.auslander_reiten_quiver()
            sage: G = AR.digraph_preprojectives(2)
            sage: G = AR._digraph_set_latex_options(G)
            sage: G.latex_options().get_option('edge_labels')
            True
        """
        G.set_latex_options(edge_labels=True)

        from sage.graphs.dot2tex_utils import have_dot2tex
        if have_dot2tex():
            from sage.misc.latex import LatexExpr

            def edge_options(data):
                u, v, l = data
                edge_opts = {}
                if l == 'ART':
                    edge_opts["color"] = "dashed,blue"
                    edge_opts["label"] = LatexExpr(r"\tau")
                return edge_opts

            G.set_latex_options(format='dot2tex', edge_options=edge_options)
        return G

    def digraph_preprojectives(self, max_depth, with_translations=False):
        r"""
        Return the digraph of preprojectives of ``self`` up to ``max_depth``.

        EXAMPLES::

            sage: D = DiGraph([[1,2,'a'], [1,2,'b']], multiedges=True)
            sage: AR = D.auslander_reiten_quiver()
            sage: G = AR.digraph_preprojectives(3)
            sage: [node.dimension_vector() for node in G]
            [v1 + 2*v2, v2, 3*v1 + 4*v2, 2*v1 + 3*v2, 5*v1 + 6*v2, 4*v1 + 5*v2]
            sage: AR.digraph_preprojectives(0)
            Digraph on 0 vertices
        """
        if max_depth < 1:
            return self._digraph_set_latex_options(DiGraph())

        k = 2
        prev = dict(self.projectives())
        verts = list(prev.values())
        edges = [(prev[v], prev[u], l) for u, v, l in self._quiver.edge_iterator()]
        cur = self._dim_vecs_level(k)
        while k <= max_depth:
            # convert cur to the appropriate data
            cur = {v: self.element_class(self, v, k) for v in cur}
            verts.extend(cur.values())
            edges.extend((cur[v], cur[u], l)
                         for u in cur for _, v, l in self._quiver.outgoing_edge_iterator(u) if v in cur)
            edges.extend((prev[u], cur[v], l) for v in cur
                         for u, _, l in self._quiver.incoming_edge_iterator(v) if u in prev)
            if with_translations:
                edges.extend((cur[v], prev[v], 'ART') for v in cur if v in prev)
            k += 1
            prev = cur
            cur = self._dim_vecs_level(k)

        G = DiGraph([verts, edges], format='vertices_and_edges', multiedges=True, immutable=True)
        return self._digraph_set_latex_options(G)

    def digraph_postinjectives(self, max_depth, with_translations=False):
        """
        Return the digraph of postinjectives of ``self`` up to ``max_depth``.

        EXAMPLES::

            sage: D = DiGraph([[1,2,'a'], [1,2,'b']], multiedges=True)
            sage: AR = D.auslander_reiten_quiver()
            sage: G = AR.digraph_postinjectives(3)
            sage: [node.dimension_vector() for node in G]
            [5*v1 + 4*v2, 6*v1 + 5*v2, 3*v1 + 2*v2, 4*v1 + 3*v2, v1, 2*v1 + v2]
            sage: AR.digraph_postinjectives(0)
            Digraph on 0 vertices
        """
        if max_depth < 1:
            return self._digraph_set_latex_options(DiGraph())

        k = 2
        prev = dict(self.injectives())
        verts = list(prev.values())
        edges = [(prev[u], prev[v], l) for u, v, l in self._quiver.edge_iterator()]
        cur = self._dim_vecs_level(-k)
        while k <= max_depth:
            # convert cur to the appropriate data
            cur = {v: self.element_class(self, v, -k) for v in cur}
            verts.extend(cur.values())
            edges.extend((cur[u], cur[v], l)
                         for u in cur for _, v, l in self._quiver.outgoing_edge_iterator(u) if v in cur)
            edges.extend((cur[v], prev[u], l) for v in cur
                         for u, _, l in self._quiver.incoming_edge_iterator(v) if u in prev)
            if with_translations:
                edges.extend((prev[v], cur[v], 'ART') for v in cur if v in prev)
            k += 1
            prev = cur
            cur = self._dim_vecs_level(-k)

        G = DiGraph([verts, edges], format='vertices_and_edges', multiedges=True, immutable=True)
        return self._digraph_set_latex_options(G)

    @cached_method
    def digraph(self, with_translations=False):
        r"""
        Return the digraph of ``self``.

        INPUT:

        - ``with_translations`` -- boolean (default: ``False``); if ``True``,
          then include the arrows corresponding to the translations

        EXAMPLES::

            sage: DA = DiGraph([[1,2]])
            sage: AR = DA.auslander_reiten_quiver()
            sage: G = AR.digraph(); G
            Digraph on 3 vertices
            sage: G.edges()
            [(<1, 1>, <2, 2>, None), (<2, 1>, <1, 1>, None)]
            sage: GT = AR.digraph(with_translations=True)
            sage: GT.edges()
            [(<1, 1>, <2, 2>, None), (<2, 1>, <1, 1>, None), (<2, 2>, <2, 1>, 'ART')]
        """
        if not self._is_finite:
            raise TypeError("the AR quiver is not finite")

        if with_translations:
            G = self.digraph().copy(immutable=False)
            for v in G.vertex_iterator():
                u = v.translation()
                if u is not None:
                    G.add_edge(v, u, 'ART')
            G = G.copy(immutable=True)

        else:
            k = 2
            prev = dict(self.projectives())
            injectives = dict(prev)  # make a shallow copy since we will mutate it
            verts = list(prev.values())
            edges = [(prev[v], prev[u], l) for u, v, l in self._quiver.edge_iterator()]
            cur = self._dim_vecs_level(k)
            while cur:
                # convert cur to the appropriate data
                cur = {v: self.element_class(self, v, k) for v in cur}
                injectives.update(cur)
                verts.extend(cur.values())
                edges.extend((cur[v], cur[u], l)
                             for u in cur for _, v, l in self._quiver.outgoing_edge_iterator(u) if v in cur)
                edges.extend((prev[u], cur[v], l) for v in cur
                             for u, _, l in self._quiver.incoming_edge_iterator(v) if u in prev)
                k += 1
                prev = cur
                cur = self._dim_vecs_level(k)

            self._injectives = Family(injectives)
            G = DiGraph([verts, edges], format='vertices_and_edges', immutable=True)

        return self._digraph_set_latex_options(G)

    def __iter__(self):
        r"""
        Iterate over ``self`` when possible.

        EXAMPLES::

            sage: DD = DiGraph([[3,2], [4,2], [2,1]])
            sage: AR = DD.auslander_reiten_quiver()
            sage: list(AR)
            [<1, 1>, <2, 1>, <3, 1>, <4, 1>, <1, 2>, <2, 2>, <3, 2>, <4, 2>,
             <1, 3>, <2, 3>, <3, 3>, <4, 3>]
        """
        return iter(self.digraph())

    def _element_constructor_(self, vertex, level=None):
        r"""
        Construct an element of ``self``.

        EXAMPLES::

            sage: DA = DiGraph([[4,3], [3,2], [2,1]])
            sage: AR = DA.auslander_reiten_quiver()
            sage: AR(2, 2)
            <2, 2>
            sage: AR((2, 3))
            <2, 3>
            sage: AR(2, 4)
            Traceback (most recent call last):
            ...
            ValueError: no 2 at level 4
            sage: AR(2, 1/2)
            Traceback (most recent call last):
            ...
            ValueError: the level 1/2 must be an integer
            sage: AR(10, 1)
            Traceback (most recent call last):
            ...
            ValueError: 10 is not a vertex of the quiver
        """
        if level is None:
            if len(vertex) == 2:
                vertex, level = vertex
        if vertex not in self._quiver:
            raise ValueError(f"{vertex} is not a vertex of the quiver")
        if level == 1:
            return self.element_class(self, vertex, level)
        if level not in ZZ:
            raise ValueError(f"the level {level} must be an integer")

        if not self._is_finite:
            if level == -1 or vertex in self._dim_vecs_level(level):
                return self.element_class(self, vertex, level)
            # This is likely never true
            raise ValueError(f"no {vertex} at level {level}")

        # otherwise the AR quiver is finite
        if level < 0:
            self.digraph()  # computes the max level
            level = self._max_level - level
        if level > 1:
            if vertex in self._dim_vecs_level(level):
                return self.element_class(self, vertex, level)
            raise ValueError(f"no {vertex} at level {level}")

    @cached_method
    def _dim_vecs_level(self, k):
        r"""
        Return a ``dict`` of dimension vectors of level ``k``.

        .. WARNING::

            This is only meant to be used internally as the output is
            mutable but cached. Thus, the output should not be changed.

        EXAMPLES::

            sage: DA = DiGraph([[4,3], [3,2], [2,1]])
            sage: AR = DA.auslander_reiten_quiver()
            sage: AR._dim_vecs_level(1)
            {1: v1, 2: v1 + v2, 3: v1 + v2 + v3, 4: v1 + v2 + v3 + v4}
            sage: AR._dim_vecs_level(2)
            {1: v2, 2: v2 + v3, 3: v2 + v3 + v4}
            sage: AR._dim_vecs_level(3)
            {1: v3, 2: v3 + v4}
            sage: AR._dim_vecs_level(4)
            {1: v4}
            sage: AR._dim_vecs_level(-1)
            {1: v1 + v2 + v3 + v4, 2: v2 + v3 + v4, 3: v3 + v4, 4: v4}
            sage: AR._dim_vecs_level(-2)
            {2: v1 + v2 + v3, 3: v2 + v3, 4: v3}
            sage: AR._dim_vecs_level(-3)
            {3: v1 + v2, 4: v2}
            sage: AR._dim_vecs_level(-4)
            {4: v1}
        """
        if k == 0:
            raise ValueError("k must not be 0")
        M = self._dim_vec_space
        Q = self._quiver
        if k == 1:
            ret = {v: M._from_dict({u: ZZ(len(Q.all_paths(v, u, use_multiedges=True))) for u in Q.vertex_iterator()})
                   for v in Q.vertex_iterator()}
        elif k > 1:
            if k > self._max_level:
                return {}
            prev = self._dim_vecs_level(k - 1)
            if k > self._max_level:  # this might get set on the recursive call
                return {}
            ret = {}
            for v in reversed(self._top_sort):
                if v not in prev:  # assumption: this vertex will never reappear
                    continue
                temp = -prev[v]
                for u, _, _ in Q.incoming_edge_iterator(v):
                    if u in prev:
                        temp += prev[u]
                for _, u, _ in Q.outgoing_edge_iterator(v):
                    if u in ret:
                        temp += ret[u]
                if all(coeff > 0 for key, coeff in temp):
                    ret[v] = temp
            if not ret:
                self._max_level = k

        elif k == -1:
            ret = {v: M._from_dict({u: ZZ(len(Q.all_paths(u, v, use_multiedges=True))) for u in Q.vertex_iterator()})
                   for v in Q.vertex_iterator()}

        elif k < -1:
            prev = self._dim_vecs_level(k + 1)
            ret = {}
            for v in self._top_sort:
                if v not in prev:  # assumption: this vertex will never reappear
                    continue
                temp = -prev[v]
                for _, u, _ in Q.outgoing_edge_iterator(v):
                    if u in prev:
                        temp += prev[u]
                for u, _, _ in Q.incoming_edge_iterator(v):
                    if u in ret:
                        temp += ret[u]
                if all(coeff > 0 for key, coeff in temp):
                    ret[v] = temp

        return ret

    def dimension_vectors_of_level(self, k):
        r"""
        Return a :class:`Family` of dimension vectors of level ``k``.

        EXAMPLES::

            sage: DA = DiGraph([[4,3], [2,3], [2,1]])
            sage: AR = DA.auslander_reiten_quiver()
            sage: AR.dimension_vectors_of_level(1)
            {1: v1, 2: v1 + v2 + v3, 3: v3, 4: v3 + v4}
            sage: AR.dimension_vectors_of_level(3)
            {1: v4, 3: v2}
            sage: AR.dimension_vectors_of_level(10)
            {}
            sage: AR.dimension_vectors_of_level(-1)
             {1: v1 + v2, 2: v2, 3: v2 + v3 + v4, 4: v4}
            sage: AR.dimension_vectors_of_level(-2)
            {1: v3 + v4, 2: v1 + v2 + v3 + v4, 3: v1 + v2 + v3, 4: v2 + v3}

            sage: D = DiGraph([[1,2,'a'], [1,2,'b']], multiedges=True)
            sage: AR = D.auslander_reiten_quiver()
            sage: AR.dimension_vectors_of_level(1)
            {1: v1 + 2*v2, 2: v2}
            sage: AR.dimension_vectors_of_level(3)
            {1: 5*v1 + 6*v2, 2: 4*v1 + 5*v2}
            sage: AR.dimension_vectors_of_level(-1)
            {1: v1, 2: 2*v1 + v2}
            sage: AR.dimension_vectors_of_level(-3)
            {1: 5*v1 + 4*v2, 2: 6*v1 + 5*v2}
        """
        ret = self._dim_vecs_level(k)
        return dict(ret)  # make a (shallow) copy to allow a user to mutate it

    class Element(Element):
        r"""
        A node in the AR quiver.
        """
        def __init__(self, parent, vertex, level):
            r"""
            Initialize ``self``.

            EXAMPLES::

                sage: DA = DiGraph([[4,3], [3,2], [2,1]])
                sage: AR = DA.auslander_reiten_quiver()
                sage: TestSuite(AR(1, 3)).run()

                sage: D = DiGraph([[1,2,'a'], [1,2,'b']], multiedges=True)
                sage: AR = D.auslander_reiten_quiver()
                sage: TestSuite(AR(2, 3)).run()
                sage: TestSuite(AR(1, -4)).run()
            """
            self._vertex = vertex
            self._level = ZZ(level)
            Element.__init__(self, parent)

        def _repr_(self) -> str:
            r"""
            Return a string representation of ``self``.

            EXAMPLES::

                sage: DA = DiGraph([[4,3], [3,2], [2,1]])
                sage: AR = DA.auslander_reiten_quiver()
                sage: AR(1, 3)
                <1, 3>
            """
            return f"<{self._vertex}, {self._level}>"

        def _latex_(self) -> str:
            r"""
            Return a latex representation of ``self``.

            EXAMPLES::

                sage: DA = DiGraph([[4,3], [3,2], [2,1]])
                sage: AR = DA.auslander_reiten_quiver()
                sage: node = AR(2, 2)
                sage: latex(node)
                \left\langle 2, 2 \right\rangle
                sage: AR.options.latex = "dimension_vector"
                sage: latex(node)
                v_{2} + v_{3}
                sage: AR.options.latex = "both"
                sage: latex(node)
                \begin{gathered} \left\langle 2, 2 \right\rangle \\ v_{2} + v_{3} \end{gathered}
                sage: AR.options._reset()
            """
            from sage.misc.latex import latex
            node = r"\left\langle {}, {} \right\rangle".format(latex(self._vertex), self._level)
            latex_option = self.parent().options.latex
            if latex_option == "node":
                return node
            dim_vec = latex(self.dimension_vector())
            if latex_option == "dimension_vector":
                return dim_vec
            return r"\begin{{gathered}} {} \\ {} \end{{gathered}}".format(node, dim_vec)

        def _richcmp_(self, other, op) -> bool:
            r"""
            Rich comparison of ``self`` to ``other`` by ``op``.

            EXAMPLES::

                sage: DA = DiGraph([[2,3], [2,1]])
                sage: AR = DA.auslander_reiten_quiver()
                sage: sorted(AR)
                [<1, 1>, <2, 1>, <3, 1>, <1, 2>, <2, 2>, <3, 2>]
            """
            return richcmp((self._level, self._vertex), (other._level, other._vertex), op)

        def __hash__(self) -> int:
            r"""
            Return the hash of ``self``.

            EXAMPLES::

                sage: DA = DiGraph([[2,3], [2,1]])
                sage: AR = DA.auslander_reiten_quiver()
                sage: node = AR(1, 2)
                sage: hash(node) == hash((2, 1))
                True
            """
            return hash((self._level, self._vertex))

        def vertex(self):
            r"""
            Return the vertex of the quiver corresponding to ``self``.

            EXAMPLES::

                sage: DA = DiGraph([[2,3], [2,1]])
                sage: AR = DA.auslander_reiten_quiver()
                sage: node = AR(1, 2)
                sage: node.vertex()
                1
            """
            return self._vertex

        def level(self):
            r"""
            Return the level of ``self``.

            EXAMPLES::

                sage: DA = DiGraph([[2,3], [2,1]])
                sage: AR = DA.auslander_reiten_quiver()
                sage: node = AR(1, 2)
                sage: node.level()
                2
            """
            return self._level

        def translation(self):
            r"""
            Return the AR translation of ``self``.

            EXAMPLES::

                sage: DA = DiGraph([[4,3], [3,2], [2,1]])
                sage: AR = DA.auslander_reiten_quiver()
                sage: node = AR(1, 1)
                sage: node.translation() is None
                True
                sage: node = AR(1, 2)
                sage: node.translation()
                <1, 1>
            """
            if self._level == 1:
                return None
            dim_vecs = self.parent()._dim_vecs_level(self._level - 1)
            if self._vertex not in dim_vecs:  # this likely never happens
                return None
            return type(self)(self.parent(), self._vertex, self._level - 1)

        def inverse_translation(self):
            r"""
            Return the inverse AR translation of ``self``.

            EXAMPLES::

                sage: DA = DiGraph([[2,3], [2,1]])
                sage: AR = DA.auslander_reiten_quiver()
                sage: node = AR(1, 1)
                sage: node.inverse_translation()
                <1, 2>
                sage: node = AR(1, 2)
                sage: node.inverse_translation() is None
                True

                sage: D = DiGraph([[1,2,'a'], [1,2,'b']], multiedges=True)
                sage: AR = D.auslander_reiten_quiver()
                sage: AR(2, -1).inverse_translation() is None
                True
            """
            if self._level == -1:
                return None
            dim_vecs = self.parent()._dim_vecs_level(self._level + 1)
            if self._vertex not in dim_vecs:
                return None
            return type(self)(self.parent(), self._vertex, self._level + 1)

        def dimension_vector(self):
            r"""
            Return the dimension vector of ``self``.

            EXAMPLES::

                sage: D = DiGraph([[1,2,'a'], [1,2,'b']], multiedges=True)
                sage: AR = D.auslander_reiten_quiver()
                sage: node = AR(2, -4)
                sage: node.dimension_vector()
                8*v1 + 7*v2
            """
            return self.parent()._dim_vecs_level(self._level)[self._vertex]


def detect_dynkin_quiver(quiver):
    """
    Determine if ``quiver`` is a finite type Dynkin quiver.

    EXAMPLES::

        sage: from sage.quivers.ar_quiver import detect_dynkin_quiver
        sage: D = DiGraph([[1,2], [2,3], [3, 4], [4,0], ['a','b'], ['b','c'], ['c','d'], ['c','e']])
        sage: detect_dynkin_quiver(D)
        D5xA5

        sage: D = DiGraph([[1,2,'a'], [1,2,'b']], multiedges=True)
        sage: detect_dynkin_quiver(D) is None
        True
        sage: D = DiGraph([[1, 2], [2, 3], [1, 3]])
        sage: detect_dynkin_quiver(D) is None
        True
        sage: D = DiGraph([[1,2], [1,3], [1,4], [1,5]])
        sage: detect_dynkin_quiver(D) is None
        True
        sage: D = DiGraph([[1,2], [2,3], [2,4], [4,5], [6,4]])
        sage: detect_dynkin_quiver(D) is None
        True
        sage: D = DiGraph([[1,2], [2,3], [3,4], [4,5], [5,6], [6,7], [7,8], [8,9], [0,3]])
        sage: detect_dynkin_quiver(D) is None
        True
        sage: D = DiGraph([[1,2], [2,3], [3,4], [4,5], [5,6], [6,7], [0,4]])
        sage: detect_dynkin_quiver(D) is None
        True
    """
    dynkin_type = []
    for Q in quiver.connected_components_subgraphs():
        if Q.has_multiple_edges():
            return None
        G = Q.to_undirected()
        if G.is_path():
            dynkin_type.append(['A', Q.n_vertices()])
            continue
        degthree = G.vertices(degree=3)
        if len(degthree) != 1:
            return None
        G = G.copy(immutable=False)
        G.delete_vertex(degthree[0])
        path_lengths = sorted(G.connected_components_sizes())
        if len(path_lengths) != 3:
            return None
        if path_lengths[:2] == [1, 1]:
            dynkin_type.append(['D', G.n_vertices() + 1])
        elif path_lengths[:2] == [1, 2] and path_lengths[2] in [2, 3, 4]:
            dynkin_type.append(['E', G.n_vertices() + 1])
        else:
            return None
    if len(dynkin_type) == 1:
        return CartanType(dynkin_type[0])
    return CartanType(dynkin_type)



# sage_setup: distribution = sagemath-objects
# distutils: libraries = gmp

cdef extern from "gmp.h":
    void mp_set_memory_functions(
        void *(*) (size_t),
        void *(*) (void *, size_t, size_t),
        void (*) (void *, size_t))



# sage_setup: distribution = sagemath-modules
"""
Root systems

Quickref
--------

- ``T = CartanType(["A", 3]), T.is_finite()``     -- Cartan types
- ``T.dynkin_diagram(), DynkinDiagram(["G",2])``  -- Dynkin diagrams
- ``T.cartan_matrix(),  CartanMatrix(["F",4])``   -- Cartan matrices
- ``RootSystem(T).weight_lattice()``              -- Root systems
- ``WeylGroup(["B", 6, 1]).simple_reflections()`` -- Affine Weyl groups
- ``WeylCharacterRing(["D", 4])``                 -- Weyl character rings

Introductory material
---------------------

- :ref:`sage.combinat.root_system.all`            -- This overview
- :class:`CartanType`                             -- An introduction to Cartan types
- :class:`RootSystem`                             -- An introduction to root systems
- :ref:`sage.combinat.root_system.plot`           -- A root system visualization tutorial

- The `Lie Methods and Related Combinatorics <../../../../../thematic_tutorials/lie.html>`_ thematic tutorial


Related material
----------------

- :ref:`sage.combinat.crystals.all`               -- Crystals

Cartan datum
------------

- :ref:`sage.combinat.root_system.cartan_type`
- :ref:`sage.combinat.root_system.dynkin_diagram`
- :ref:`sage.combinat.root_system.cartan_matrix`
- :ref:`sage.combinat.root_system.coxeter_matrix`
- :ref:`sage.combinat.root_system.coxeter_type`

Root systems
------------

- :ref:`sage.combinat.root_system.all`
- :ref:`sage.combinat.root_system.plot`
- :ref:`sage.combinat.root_system.root_lattice_realizations`
- :ref:`sage.combinat.root_system.root_lattice_realization_algebras`
- :ref:`sage.combinat.root_system.weight_lattice_realizations`
- :ref:`sage.combinat.root_system.root_space`
- :ref:`sage.combinat.root_system.weight_space`
- :ref:`sage.combinat.root_system.ambient_space`

Coxeter groups
--------------

- :ref:`sage.combinat.root_system.coxeter_group`
- :ref:`sage.combinat.root_system.weyl_group`
- :ref:`sage.combinat.root_system.extended_affine_weyl_group`
- :ref:`sage.combinat.root_system.fundamental_group`
- :ref:`sage.combinat.root_system.braid_move_calculator`
- :ref:`sage.combinat.root_system.braid_orbit`

.. SEEALSO::

    The categories :class:`CoxeterGroups` and :class:`WeylGroups`

Finite reflection groups
------------------------

- :ref:`sage.combinat.root_system.reflection_group_complex`
- :ref:`sage.combinat.root_system.reflection_group_real`

.. SEEALSO::

    The category :class:`~sage.categories.complex_reflection_groups.ComplexReflectionGroups`

Representation theory
---------------------

- :ref:`sage.combinat.root_system.weyl_characters`
- :ref:`sage.algebras.fusion_rings.fusion_ring`
- :ref:`sage.combinat.root_system.integrable_representations`
- :ref:`sage.combinat.root_system.branching_rules`
- :ref:`sage.combinat.root_system.hecke_algebra_representation`
- :ref:`sage.combinat.root_system.non_symmetric_macdonald_polynomials`

Root system data and code for specific families of Cartan types
---------------------------------------------------------------

- :ref:`sage.combinat.root_system.type_affine`
- :ref:`sage.combinat.root_system.type_dual`
- :ref:`sage.combinat.root_system.type_folded`
- :ref:`sage.combinat.root_system.type_reducible`
- :ref:`sage.combinat.root_system.type_relabel`
- :ref:`sage.combinat.root_system.type_marked`

Root system data and code for specific Cartan types
---------------------------------------------------

- :ref:`sage.combinat.root_system.type_A`
- :ref:`sage.combinat.root_system.type_B`
- :ref:`sage.combinat.root_system.type_C`
- :ref:`sage.combinat.root_system.type_D`
- :ref:`sage.combinat.root_system.type_E`
- :ref:`sage.combinat.root_system.type_F`
- :ref:`sage.combinat.root_system.type_G`
- :ref:`sage.combinat.root_system.type_H`
- :ref:`sage.combinat.root_system.type_I`
- :ref:`sage.combinat.root_system.type_A_affine`
- :ref:`sage.combinat.root_system.type_B_affine`
- :ref:`sage.combinat.root_system.type_C_affine`
- :ref:`sage.combinat.root_system.type_D_affine`
- :ref:`sage.combinat.root_system.type_E_affine`
- :ref:`sage.combinat.root_system.type_F_affine`
- :ref:`sage.combinat.root_system.type_G_affine`
- :ref:`sage.combinat.root_system.type_BC_affine`
- :ref:`sage.combinat.root_system.type_super_A`
- :ref:`sage.combinat.root_system.type_A_infinity`
"""
# install the docstring of this module to the containing package
from sage.misc.namespace_package import install_doc
install_doc(__package__, __doc__)

from sage.misc.lazy_import import lazy_import

from sage.combinat.root_system.cartan_type import CartanType
lazy_import('sage.combinat.root_system.dynkin_diagram', 'DynkinDiagram')
lazy_import('sage.combinat.root_system.cartan_matrix', 'CartanMatrix')
lazy_import('sage.combinat.root_system.coxeter_matrix', 'CoxeterMatrix')
from sage.combinat.root_system.coxeter_type import CoxeterType
from sage.combinat.root_system.root_system import RootSystem, WeylDim
lazy_import('sage.combinat.root_system.weyl_group', ['WeylGroup',
                                                     'WeylGroupElement'])
lazy_import('sage.combinat.root_system.reflection_group_real',
            'ReflectionGroup')
lazy_import('sage.combinat.root_system.extended_affine_weyl_group',
            'ExtendedAffineWeylGroup')
lazy_import('sage.combinat.root_system.coxeter_group', 'CoxeterGroup')
lazy_import('sage.combinat.root_system.weyl_characters', ['WeylCharacterRing',
                                                          'WeightRing'])
from sage.combinat.root_system.branching_rules import BranchingRule, branching_rule_from_plethysm, branching_rule

lazy_import('sage.combinat.root_system.non_symmetric_macdonald_polynomials',
            'NonSymmetricMacdonaldPolynomials')
lazy_import('sage.combinat.root_system.integrable_representations',
            'IntegrableRepresentation')
del lazy_import
del install_doc



# sage_setup: distribution = sagemath-glpk
#*****************************************************************************
#       Copyright (C) 2012 Christian Kuper <christian.kuper@t-online.de>
#       Copyright (C) 2015 Jeroen Demeyer <jdemeyer@cage.ugent.be>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.libs.glpk.types cimport glp_graph

ctypedef struct c_v_data:
         double rhs
         double pi
         double es
         double ls
         long cut

ctypedef struct c_a_data:
         double low
         double cap
         double cost
         double x


cdef class GLPKGraphBackend():
    cdef glp_graph * graph
    cpdef add_vertex(self, name=?)
    cpdef list add_vertices(self, vertices)
    cpdef __add_vertices_sage(self, g)
    cpdef dict get_vertex(self, vertex)
    cpdef dict get_vertices(self, verts)
    cpdef set_vertex_demand(self, vertex, param)
    cpdef set_vertices_demand(self, list pairs)
    cpdef list vertices(self)
    cpdef add_edge(self, u, v, dict params=?)
    cpdef __add_edges_sage(self, g)
    cpdef list add_edges(self, edges)
    cpdef delete_edge(self, u, v, dict params=?)
    cpdef tuple get_edge(self, u, v)
    cpdef list edges(self)
    cpdef delete_vertex(self, vert)
    cpdef delete_vertices(self, list verts)
    cpdef int _find_vertex(self, vert) noexcept
    cpdef int write_graph(self, fname) noexcept
    cpdef int write_ccdata(self, fname) noexcept
    cpdef int write_mincost(self, fname) noexcept
    cpdef double mincost_okalg(self) except -1
    cdef int s
    cdef int t
    cpdef int write_maxflow(self, fname) except -1
    cpdef double maxflow_ffalg(self, u=?, v=?) except -1
    cpdef double cpp(self) noexcept
